#!/usr/bin/env python
"""
Script that dumps the DB information for the elements into the standard output.
If returns information concerning the StatusType and Status attributes.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import datetime
from DIRAC import gLogger, exit as DIRACExit, S_OK, version
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.ResourceStatusSystem.Client import ResourceStatusClient
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Utilities.PrettyPrint import printTable

subLogger = None


def registerSwitches():
    """
    Registers all switches that can be used while calling the script from the
    command line interface.
    """

    switches = (
        ("element=", "Element family to be Synchronized ( Site, Resource, Node )"),
        ("tableType=", "A valid table type (Status, Log, History)"),
        ("name=", "ElementName (comma separated list allowed); None if default"),
        (
            "statusType=",
            "A valid StatusType argument (it admits a comma-separated list of statusTypes); None if default",
        ),
        ("status=", "A valid Status argument ( Active, Probing, Degraded, Banned, Unknown, Error ); None if default"),
        ("elementType=", "ElementType narrows the search; None if default"),
        ("reason=", "Decision that triggered the assigned status"),
        ("lastCheckTime=", "Time-stamp setting last time the status & status were checked"),
        ("tokenOwner=", "Owner of the token ( to specify only with select/delete queries"),
        ("VO=", "Virtual organisation; None if default"),
    )

    for switch in switches:
        Script.registerSwitch("", switch[0], switch[1])


def registerUsageMessage():
    """
    Takes the script __doc__ and adds the DIRAC version to it
    """

    usageMessage = "DIRAC version: %s \n" % version
    usageMessage += __doc__

    Script.setUsageMessage(usageMessage)


def parseSwitches():
    """
    Parses the arguments passed by the user
    """

    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()
    if len(args) < 3:
        error("Missing all mandatory 'query', 'element', 'tableType' arguments")
    elif args[0].lower() not in ("select", "add", "modify", "delete"):
        error("Incorrect 'query' argument")
    elif args[1].lower() not in ("site", "resource", "component", "node"):
        error("Incorrect 'element' argument")
    elif args[2].lower() not in ("status", "log", "history"):
        error("Incorrect 'tableType' argument")
    else:
        query = args[0].lower()

    switches = dict(Script.getUnprocessedSwitches())

    # Default values
    switches.setdefault("name", None)
    switches.setdefault("statusType", None)
    switches.setdefault("status", None)
    switches.setdefault("elementType", None)
    switches.setdefault("reason", None)
    switches.setdefault("lastCheckTime", None)
    switches.setdefault("tokenOwner", None)
    switches.setdefault("VO", None)

    if "status" in switches and switches["status"] is not None:
        switches["status"] = switches["status"].title()
        if switches["status"] not in ("Active", "Probing", "Degraded", "Banned", "Unknown", "Error"):
            error("'%s' is an invalid argument for switch 'status'" % switches["status"])

    # when it's a add/modify query and status/reason/statusType are not specified
    # then some specific defaults are set up
    if query == "add" or query == "modify":
        if "status" not in switches or switches["status"] is None:
            switches["status"] = "Unknown"
        if "reason" not in switches or switches["reason"] is None:
            switches["reason"] = "Unknown reason"
        if "statusType" not in switches or switches["statusType"] is None:
            switches["statusType"] = "all"

    subLogger.debug("The switches used are:")
    map(subLogger.debug, switches.items())

    return args, switches


# UTILS: to check and unpack


def getToken(key):
    """
    Function that gets the userName from the proxy
    """

    proxyInfo = getProxyInfo()
    if not proxyInfo["OK"]:
        error(str(proxyInfo))

    if key.lower() == "owner":
        userName = proxyInfo["Value"]["username"]
        tokenOwner = S_OK(userName)
        if not tokenOwner["OK"]:
            error(tokenOwner["Message"])
        return tokenOwner["Value"]

    elif key.lower() == "expiration":
        expiration = proxyInfo["Value"]["secondsLeft"]
        tokenExpiration = S_OK(expiration)
        if not tokenExpiration["OK"]:
            error(tokenExpiration["Message"])

        now = Time.dateTime()
        # datetime.datetime.utcnow()
        expirationDate = now + datetime.timedelta(seconds=tokenExpiration["Value"])
        return expirationDate


def checkStatusTypes(statusTypes):
    """
    To check if values for 'statusType' are valid
    """

    opsH = Operations().getValue(
        "ResourceStatus/Config/StatusTypes/StorageElement", "ReadAccess,WriteAccess,CheckAccess,RemoveAccess"
    )
    acceptableStatusTypes = opsH.replace(",", "").split()

    for statusType in statusTypes:
        if statusType not in acceptableStatusTypes and statusType != "all":
            acceptableStatusTypes.append("all")
            error(
                "'%s' is a wrong value for switch 'statusType'.\n\tThe acceptable values are:\n\t%s"
                % (statusType, str(acceptableStatusTypes))
            )

    if "all" in statusType:
        return acceptableStatusTypes
    return statusTypes


def unpack(switchDict):
    """
    To split and process comma-separated list of values for 'name' and 'statusType'
    """

    switchDictSet = []
    names = []
    statusTypes = []

    if switchDict["name"] is not None:
        names = list(filter(None, switchDict["name"].split(",")))

    if switchDict["statusType"] is not None:
        statusTypes = list(filter(None, switchDict["statusType"].split(",")))
        statusTypes = checkStatusTypes(statusTypes)

    if names and statusTypes:
        combinations = [(a, b) for a in names for b in statusTypes]
        for combination in combinations:
            n, s = combination
            switchDictClone = switchDict.copy()
            switchDictClone["name"] = n
            switchDictClone["statusType"] = s
            switchDictSet.append(switchDictClone)
    elif names and not statusTypes:
        for name in names:
            switchDictClone = switchDict.copy()
            switchDictClone["name"] = name
            switchDictSet.append(switchDictClone)
    elif not names and statusTypes:
        for statusType in statusTypes:
            switchDictClone = switchDict.copy()
            switchDictClone["statusType"] = statusType
            switchDictSet.append(switchDictClone)
    elif not names and not statusTypes:
        switchDictClone = switchDict.copy()
        switchDictClone["name"] = None
        switchDictClone["statusType"] = None
        switchDictSet.append(switchDictClone)

    return switchDictSet


# UTILS: for filtering 'select' output


def filterReason(selectOutput, reason):
    """
    Selects all the elements that match 'reason'
    """

    elements = selectOutput
    elementsFiltered = []
    if reason is not None:
        for e in elements:
            if reason in e["reason"]:
                elementsFiltered.append(e)
    else:
        elementsFiltered = elements

    return elementsFiltered


# Utils: for formatting query output and notifications


def error(msg):
    """
    Format error messages
    """

    subLogger.error("\nERROR:")
    subLogger.error("\t" + msg)
    subLogger.error("\tPlease, check documentation below")
    Script.showHelp(exitCode=1)


def confirm(query, matches):
    """
    Format confirmation messages
    """

    subLogger.notice("\nNOTICE: '%s' request successfully executed ( matches' number: %s )! \n" % (query, matches))


def tabularPrint(table):

    columns_names = list(table[0])
    records = []
    for row in table:
        record = []
        for _k, v in row.items():
            if isinstance(v, datetime.datetime):
                record.append(Time.toString(v))
            elif v is None:
                record.append("")
            else:
                record.append(v)
        records.append(record)

    output = printTable(columns_names, records, numbering=False, columnSeparator=" | ", printOut=False)

    subLogger.notice(output)


def select(args, switchDict):
    """
    Given the switches, request a query 'select' on the ResourceStatusDB
    that gets from <element><tableType> all rows that match the parameters given.
    """

    rssClient = ResourceStatusClient.ResourceStatusClient()

    meta = {
        "columns": [
            "name",
            "statusType",
            "status",
            "elementType",
            "reason",
            "dateEffective",
            "lastCheckTime",
            "tokenOwner",
            "tokenExpiration",
            "vO",
        ]
    }

    result = {"output": None, "successful": None, "message": None, "match": None}
    output = rssClient.selectStatusElement(
        element=args[1].title(),
        tableType=args[2].title(),
        name=switchDict["name"],
        statusType=switchDict["statusType"],
        status=switchDict["status"],
        elementType=switchDict["elementType"],
        lastCheckTime=switchDict["lastCheckTime"],
        tokenOwner=switchDict["tokenOwner"],
        vO=switchDict["VO"],
        meta=meta,
    )
    result["output"] = [dict(zip(output["Columns"], e)) for e in output["Value"]]
    result["output"] = filterReason(result["output"], switchDict["reason"])
    result["match"] = len(result["output"])
    result["successful"] = output["OK"]
    result["message"] = output["Message"] if "Message" in output else None

    return result


def add(args, switchDict):
    """
    Given the switches, request a query 'addOrModify' on the ResourceStatusDB
    that inserts or updates-if-duplicated from <element><tableType> and also adds
    a log if flag is active.
    """

    rssClient = ResourceStatusClient.ResourceStatusClient()

    result = {"output": None, "successful": None, "message": None, "match": None}
    output = rssClient.addOrModifyStatusElement(
        element=args[1].title(),
        tableType=args[2].title(),
        name=switchDict["name"],
        statusType=switchDict["statusType"],
        status=switchDict["status"],
        elementType=switchDict["elementType"],
        reason=switchDict["reason"],
        tokenOwner=getToken("owner"),
        tokenExpiration=getToken("expiration"),
        vO=switchDict["VO"],
    )

    if output.get("Value"):
        result["match"] = int(output["Value"] if output["Value"] else 0)
    result["successful"] = output["OK"]
    result["message"] = output["Message"] if "Message" in output else None

    return result


def modify(args, switchDict):
    """
    Given the switches, request a query 'modify' on the ResourceStatusDB
    that updates from <element><tableType> and also adds a log if flag is active.
    """

    rssClient = ResourceStatusClient.ResourceStatusClient()

    result = {"output": None, "successful": None, "message": None, "match": None}
    output = rssClient.modifyStatusElement(
        element=args[1].title(),
        tableType=args[2].title(),
        name=switchDict["name"],
        statusType=switchDict["statusType"],
        status=switchDict["status"],
        elementType=switchDict["elementType"],
        reason=switchDict["reason"],
        tokenOwner=getToken("owner"),
        tokenExpiration=getToken("expiration"),
        vO=switchDict["VO"],
    )

    if output.get("Value"):
        result["match"] = int(output["Value"] if output["Value"] else 0)
    result["successful"] = output["OK"]
    result["message"] = output["Message"] if "Message" in output else None

    return result


def delete(args, switchDict):
    """
    Given the switches, request a query 'delete' on the ResourceStatusDB
    that deletes from <element><tableType> all rows that match the parameters given.
    """

    rssClient = ResourceStatusClient.ResourceStatusClient()

    result = {"output": None, "successful": None, "message": None, "match": None}
    output = rssClient.deleteStatusElement(
        element=args[1].title(),
        tableType=args[2].title(),
        name=switchDict["name"],
        statusType=switchDict["statusType"],
        status=switchDict["status"],
        elementType=switchDict["elementType"],
        reason=switchDict["reason"],
        tokenOwner=switchDict["tokenOwner"],
        vO=switchDict["VO"],
    )

    if "Value" in output:
        result["match"] = int(output["Value"] if output["Value"] else 0)
    result["successful"] = output["OK"]
    result["message"] = output["Message"] if "Message" in output else None

    return result


def run(args, switchDictSet):
    """
    Main function of the script
    """
    query = args[0]

    matches = 0
    table = []

    for switchDict in switchDictSet:
        # exectue the query request: e.g. if it's a 'select' it executes 'select()'
        # the same if it is insert, update, add, modify, delete
        result = eval(query + "( args, switchDict )")

        if result["successful"]:
            if query == "select" and result["match"] > 0:
                table.extend(result["output"])
            matches = matches + result["match"] if result["match"] else 0
        else:
            error(result["message"])

    if query == "select" and matches > 0:
        tabularPrint(table)
    confirm(query, matches)


@Script()
def main():
    global subLogger

    subLogger = gLogger.getSubLogger(__file__)

    # Script initialization
    registerSwitches()
    registerUsageMessage()
    args, switchDict = parseSwitches()

    # Unpack switchDict if 'name' or 'statusType' have multiple values
    switchDictSet = unpack(switchDict)

    # Run script
    run(args, switchDictSet)

    # Bye
    DIRACExit(0)


if __name__ == "__main__":
    main()



""" PilotEfficiencyPolicy

  Policy that gets efficiency from the PilotCommand result and
  sets the resource status. Efficiency is given in percent.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK
from DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase import PolicyBase

__RCSID__ = "$Id$"


class PilotEfficiencyPolicy(PolicyBase):
    """PilotEfficiencyPolicy class, extends PolicyBase"""

    @staticmethod
    def _evaluate(commandResult):
        """_evaluate

        efficiency < 50.0 :: Banned
        efficiency < 90.0 :: Degraded

        """
        # ideally,  this should be obtained from config.
        bannedLimit = 50.0
        degradedLimit = 90.0

        result = {"Status": None, "Reason": None}

        if not commandResult["OK"]:
            result["Status"] = "Error"
            result["Reason"] = commandResult["Message"]
            return S_OK(result)

        commandResult = commandResult["Value"]

        if not commandResult:
            result["Status"] = "Unknown"
            result["Reason"] = "No values to take a decision"
            return S_OK(result)

        commandResult = commandResult[0]

        if not commandResult:
            result["Status"] = "Unknown"
            result["Reason"] = "No values to take a decision"
            return S_OK(result)

        # Pilot efficiency is now available directly from the command result, in percent:
        efficiency = commandResult.get("PilotJobEff", None)
        # get the VO from the result, if present
        result["VO"] = commandResult.get("VO", None)

        if efficiency is None:
            result["Status"] = "Unknown"
            result["Reason"] = "Not enough pilots to take a decision"
            return S_OK(result)

        if efficiency <= bannedLimit:
            result["Status"] = "Banned"
        elif efficiency <= degradedLimit:
            result["Status"] = "Degraded"
        else:
            result["Status"] = "Active"

        result["Reason"] = "Pilots Efficiency of %.2f" % efficiency
        return S_OK(result)



""" ProxyProvider implementation for a per-user sub-proxy(PUSP) proxy generation using PUSP proxy server.
    More details about PUSP here: https://wiki.egi.eu/wiki/Usage_of_the_per_user_sub_proxy_in_EGI

    .. literalinclude:: /dirac.cfg
      :start-after: ## PUSP type:
      :end-before: ##
      :dedent: 2
      :caption: /Resources/ProxyProviders section
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from six.moves.urllib.request import urlopen

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.Resources.ProxyProvider.ProxyProvider import ProxyProvider

__RCSID__ = "$Id$"


class PUSPProxyProvider(ProxyProvider):
    def __init__(self, parameters=None):

        super(PUSPProxyProvider, self).__init__(parameters)

    def checkStatus(self, userDN):
        """Read ready to work status of proxy provider

        :param str userDN: user DN

        :return: S_OK()/S_ERROR()
        """
        # Search a unique identifier(username) to use as cn-label to generate PUSP
        if not userDN.split(":")[-1]:
            return S_ERROR("Can not found user label for DN: %s" % userDN)
        if not self.parameters.get("ServiceURL"):
            return S_ERROR("Can not determine PUSP service URL")

        return S_OK()

    def getProxy(self, userDN):
        """Generate user proxy

        :param str userDN: user DN

        :return: S_OK(str)/S_ERROR() -- contain a proxy string
        """
        result = self.checkStatus(userDN)
        if not result["OK"]:
            return result

        puspURL = self.parameters["ServiceURL"]
        puspURL += "?proxy-renewal=false&disable-voms-proxy=true&rfc-proxy=true"
        puspURL += "&cn-label=user:%s" % userDN.split(":")[-1]

        try:
            proxy = urlopen(puspURL).read()
        except Exception:
            return S_ERROR("Failed to get proxy from the PUSP server")

        chain = X509Chain()
        chain.loadChainFromString(proxy)
        chain.loadKeyFromString(proxy)

        result = chain.getCredentials()
        if not result["OK"]:
            return S_ERROR("Failed to get a valid PUSP proxy")
        credDict = result["Value"]
        if credDict["identity"] != userDN:
            return S_ERROR("Requested DN does not match the obtained one in the PUSP proxy")

        return chain.generateProxyToString(lifeTime=credDict["secondsLeft"])



""" Commits monitoring information using gServiceInterface singleton
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import os

from DIRAC import gLogger, gConfig, rootPath, S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities import DEncode, Time
from DIRAC.Core.Utilities.JEncode import strToIntDict
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.FrameworkSystem.private.monitoring.ServiceInterface import gServiceInterface


def initializeMonitoringHandler(serviceInfo):
    # Check that the path is writable
    monitoringSection = PathFinder.getServiceSection("Framework/Monitoring")
    # Get data location
    dataPath = gConfig.getValue("%s/DataLocation" % monitoringSection, "data/monitoring")
    dataPath = dataPath.strip()
    if dataPath[0] != "/":
        dataPath = os.path.realpath("%s/%s" % (gConfig.getValue("/LocalSite/InstancePath", rootPath), dataPath))
    gLogger.info("Data will be written into %s" % dataPath)
    try:
        os.makedirs(dataPath)
    except Exception:
        pass
    try:
        testFile = "%s/mon.jarl.test" % dataPath
        with open(testFile, "w"):
            pass
        os.unlink(testFile)
    except IOError:
        gLogger.fatal("Can't write to %s" % dataPath)
        return S_ERROR("Data location is not writable")
    # Define globals
    gServiceInterface.initialize(dataPath)
    if not gServiceInterface.initializeDB():
        return S_ERROR("Can't start db engine")
    gMonitor.registerActivity("cachedplots", "Cached plot images", "Monitoring plots", "plots", gMonitor.OP_SUM)
    gMonitor.registerActivity("drawnplots", "Drawn plot images", "Monitoring plots", "plots", gMonitor.OP_SUM)
    return S_OK()


class MonitoringHandler(RequestHandler):
    """
    This class is inherited from the RequestHandler class which is the base required for running
    any kind of services in DIRAC.
    """

    types_registerActivities = [dict, dict]

    def export_registerActivities(self, sourceDict, activitiesDict, componentExtraInfo={}):
        """
        Registers new activities
        """
        return gServiceInterface.registerActivities(sourceDict, activitiesDict, componentExtraInfo)

    types_commitMarks = [int, dict]

    # We can ignore encoding warning here because are effectively casting inside the method
    @DEncode.ignoreEncodeWarning
    def export_commitMarks(self, sourceId, activitiesDict, componentExtraInfo={}):
        """
        Adds marks for activities

        :param int sourceId: component id
        :param dict activitiesDict: the key is the metrice, the value is a dictionary with timestamp and value
                                    for example::

                                                 {u'CPU': {1583244000: 10.6055594906},
                                                  u'Iteration': {1583244420: 1,
                                                                 1583244480: 1,
                                                                 1583244540: 1,
                                                                 1583244600: 1,
                                                                 1583244660: 1},
                                                  u'MEM': {1583244000: 49.03515625},
                                                  u'Processed': {1583244420: 5,
                                                                 1583244480: 5,
                                                                 1583244540: 5,
                                                                 1583244600: 5,
                                                                 1583244660: 6}}

        :param dict componentExtraInfo: extra information
        """

        # we have to cast the dictionary keys, because they are timestamps
        activitiesDict = {metric: strToIntDict(value) for metric, value in activitiesDict.items()}
        nowEpoch = Time.toEpoch()
        maxEpoch = nowEpoch + 7200
        minEpoch = nowEpoch - 86400
        invalidActivities = []
        for acName in activitiesDict:
            for time in activitiesDict[acName]:
                if time > maxEpoch or time < minEpoch:
                    gLogger.info("Time %s  ( [%s,%s] ) is invalid for activity %s" % (time, minEpoch, maxEpoch, acName))
                    invalidActivities.append(acName)
                    break
        for acName in invalidActivities:
            gLogger.info("Not commiting activity %s" % acName)
            del activitiesDict[acName]
        return gServiceInterface.commitMarks(sourceId, activitiesDict, componentExtraInfo)

    types_queryField = [six.string_types, dict]

    def export_queryField(self, field, definedFields):
        """
        Returns available values for a field., given a set of fields and values,
        """
        definedFields["sources.setup"] = self.serviceInfoDict["clientSetup"]
        return gServiceInterface.fieldValue(field, definedFields)

    types_tryView = [int, int, six.string_types]

    def export_tryView(self, fromSecs, toSecs, viewDescriptionStub):
        """
        Generates plots based on a DEncoded view description
        """
        viewDescription, _stubLength = DEncode.decode(viewDescriptionStub)
        if "definition" not in viewDescription:
            return S_ERROR("No plot definition given")
        defDict = viewDescription["definition"]
        defDict["sources.setup"] = self.serviceInfoDict["clientSetup"]
        return gServiceInterface.generatePlots(fromSecs, toSecs, viewDescription)

    types_saveView = [six.string_types, six.string_types]

    def export_saveView(self, viewName, viewDescriptionStub):
        """
        Saves a view
        """
        if not viewName:
            return S_OK("View name not valid")
        viewDescription, _stubLength = DEncode.decode(viewDescriptionStub)
        if "definition" not in viewDescription:
            return S_ERROR("No plot definition given")
        defDict = viewDescription["definition"]
        defDict["sources.setup"] = self.serviceInfoDict["clientSetup"]
        return gServiceInterface.saveView(viewName, viewDescription)

    types_getViews = []

    def export_getViews(self, onlyStatic=True):
        """
        Returns a list of stored views
        """
        return gServiceInterface.getViews(onlyStatic)

    types_plotView = [dict]

    def export_plotView(self, viewRequest):
        """
        Generates plots for a view
        """
        for required in ("fromSecs", "toSecs", "id"):
            if required not in viewRequest:
                return S_ERROR("Missing %s field in request" % required)
        for intFields in ("fromSecs", "toSecs"):
            viewRequest[intFields] = int(viewRequest[intFields])
        if "size" not in viewRequest:
            viewRequest["size"] = 1
        if viewRequest["size"] not in (0, 1, 2, 3):
            return S_ERROR("Invalid size")
        return gServiceInterface.plotView(viewRequest)

    types_deleteView = [int]

    def export_deleteView(self, viewId):
        """
        Deletes a view
        """
        return gServiceInterface.deleteView(viewId)

    types_deleteViews = [list]

    def export_deleteViews(self, viewList):
        """
        Deletes a view
        """
        for viewId in viewList:
            result = gServiceInterface.deleteView(viewId)
            if not result["OK"]:
                return result
        return S_OK()

    types_getActivities = []

    def export_getActivities(self):
        """
        Returns a list of defined activities
        """
        dbCondition = {"sources.setup": self.serviceInfoDict["clientSetup"]}
        return S_OK(gServiceInterface.getActivities(dbCondition))

    types_getActivitiesContents = [dict, (list, tuple), six.integer_types, six.integer_types]

    def export_getActivitiesContents(self, selDict, sortList, start, limit):
        """
        Retrieve the contents of the activity db
        """
        setupCond = {"sources.setup": self.serviceInfoDict["clientSetup"]}
        selDict.update(setupCond)
        result = gServiceInterface.getActivitiesContents(selDict, sortList, start, limit)
        if not result["OK"]:
            return result
        resultTuple = result["Value"]
        result = {"Records": resultTuple[0], "Fields": resultTuple[1]}
        result["TotalRecords"] = gServiceInterface.getNumberOfActivities(setupCond)
        return S_OK(result)

    types_deleteActivity = [int, int]

    def export_deleteActivity(self, sourceId, activityId):
        """
        Deletes an activity
        """
        return gServiceInterface.deleteActivity(sourceId, activityId)

    types_deleteActivities = [list]

    def export_deleteActivities(self, deletionList):
        """
        Deletes a list of activities
        """
        failed = []
        for acList in deletionList:
            retVal = gServiceInterface.deleteActivity(acList[0], acList[1])
            if not retVal["OK"]:
                failed.append(retVal["Message"])
        if failed:
            return S_ERROR("\n".join(failed))
        return S_OK()

    types_getComponentsStatus = [dict]

    def export_getComponentsStatus(self, condDict):
        """
        Shows the status of the components.
        """
        if "Setup" not in condDict:
            condDict["Setup"] = self.serviceInfoDict["clientSetup"]
        return gServiceInterface.getComponentsStatus(condDict)

    def transfer_toClient(self, fileId, token, fileHelper):
        """
        Transfers files
        """
        retVal = gServiceInterface.getGraphData(fileId)
        if not retVal["OK"]:
            return retVal
        retVal = fileHelper.sendData(retVal["Value"])
        if not retVal["OK"]:
            return retVal
        fileHelper.sendEOF()
        return S_OK()



""" VirtualMachineMonitorAgent plays the role of the watch dog for the Virtual Machine
"""

import os
import time
import glob
import psutil

from DIRAC import S_OK, S_ERROR, gConfig, rootPath
from DIRAC.ConfigurationSystem.Client.Helpers import Operations
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities import List, Network
from DIRAC.WorkloadManagementSystem.Client.ServerUtils import virtualMachineDB

__RCSID__ = "$Id$"


class VirtualMachineMonitorAgent(AgentModule):
    def __getCSConfig(self):
        if not self.runningPod:
            return S_ERROR("/LocalSite/RunningPod is not defined")
        # Variables coming from the vm
        imgPath = "/Cloud/%s" % self.runningPod
        for csOption, csDefault, varName in (
            ("LoadAverageTimespan", 60, "vmLoadAvgTimespan"),
            ("HaltPeriod", 600, "haltPeriod"),
            ("HaltBeforeMargin", 300, "haltBeforeMargin"),
            ("HeartBeatPeriod", 300, "heartBeatPeriod"),
        ):
            path = "%s/%s" % (imgPath, csOption)
            value = self.op.getValue(path, csDefault)
            if not value > 0:
                return S_ERROR("%s has an incorrect value, must be > 0" % path)
            setattr(self, varName, value)

        for csOption, csDefault, varName in (("JobWrappersLocation", "/scratch", "vmJobWrappersLocation"),):

            path = "%s/%s" % (imgPath, csOption)
            value = gConfig.getValue(path, csDefault)
            if not value:
                return S_ERROR("%s points to an empty string, cannot be!" % path)
            setattr(self, varName, value)

        self.haltBeforeMargin = max(self.haltBeforeMargin, int(self.am_getPollingTime()) + 5)
        self.haltPeriod = max(self.haltPeriod, int(self.am_getPollingTime()) + 5)
        self.heartBeatPeriod = max(self.heartBeatPeriod, int(self.am_getPollingTime()) + 5)

        self.log.info("** VM Info **")
        self.log.info("Name                  : %s" % self.runningPod)
        self.log.info("Load Avg Timespan     : %d" % self.vmLoadAvgTimespan)
        self.log.info("Job wrappers location : %s" % self.vmJobWrappersLocation)
        self.log.info("Halt Period           : %d" % self.haltPeriod)
        self.log.info("Halt Before Margin    : %d" % self.haltBeforeMargin)
        self.log.info("HeartBeat Period      : %d" % self.heartBeatPeriod)
        if self.vmID:
            self.log.info("DIRAC ID              : %s" % self.vmID)
        if self.uniqueID:
            self.log.info("Unique ID             : %s" % self.uniqueID)
        self.log.info("*************")
        return S_OK()

    def __declareInstanceRunning(self):
        # Connect to VM monitor and register as running
        retries = 3
        sleepTime = 30
        for i in range(retries):
            result = virtualMachineDB.declareInstanceRunning(self.uniqueID, self.ipAddress)
            if result["OK"]:
                self.log.info("Declared instance running")
                return result
            self.log.error("Could not declare instance running", result["Message"])
            if i < retries - 1:
                self.log.info("Sleeping for %d seconds and retrying" % sleepTime)
                time.sleep(sleepTime)

        return S_ERROR("Could not declare instance running after %d retries" % retries)

    def initialize(self):

        self.am_disableMonitoring()
        self.op = Operations.Operations()
        # Init vars
        self.runningPod = gConfig.getValue("/LocalSite/RunningPod")
        self.log.info("Running pod name of the image is %s" % self.runningPod)
        self.vmID = gConfig.getValue("/LocalSite/VMID")

        self.__loadHistory = []
        self.vmLoadAvgTimespan = None
        self.vmJobWrappersLocation = None
        self.haltPeriod = None
        self.haltBeforeMargin = None
        self.heartBeatPeriod = None
        self.am_setOption("MaxCycles", 0)
        self.am_setOption("PollingTime", 60)

        # Discover net address
        self.ipAddress = None
        netData = Network.discoverInterfaces()
        for iface in sorted(netData):
            # Warning! On different clouds interface name may be different(eth, ens, ...)
            if "eth" in iface or "ens" in iface:
                self.ipAddress = netData[iface]["ip"]
                self.log.info("IP Address is %s" % self.ipAddress)
                break

        # Declare instance running
        self.uniqueID = ""
        result = virtualMachineDB.getUniqueIDByName(self.vmID)
        if result["OK"]:
            self.uniqueID = result["Value"]
        result = self.__declareInstanceRunning()
        if not result["OK"]:
            self.log.error("Could not declare instance running", result["Message"])
            self.__haltInstance()
            return S_ERROR("Halting!")

        self.__instanceInfo = result["Value"]

        # Get the cs config
        result = self.__getCSConfig()
        if not result["OK"]:
            return result

        return S_OK()

    def __isJobAgentRunning(self):
        isRunning = False
        for proc in psutil.process_iter(["cmdline"]):
            if "WorkloadManagement/JobAgent" in proc.info["cmdline"]:
                isRunning = True
        return isRunning

    def __getLoadAvg(self):
        result = self.__getCSConfig()
        if not result["OK"]:
            return result
        with open("/proc/loadavg", "r") as fd:
            data = [float(v) for v in List.fromChar(fd.read(), " ")[:3]]
        self.__loadHistory.append(data)
        numRequiredSamples = max(self.vmLoadAvgTimespan / self.am_getPollingTime(), 1)
        while len(self.__loadHistory) > numRequiredSamples:
            self.__loadHistory.pop(0)
        self.log.info("Load averaged over %d seconds" % self.vmLoadAvgTimespan)
        self.log.info(" %d/%s required samples to average load" % (len(self.__loadHistory), numRequiredSamples))
        avgLoad = 0
        for f in self.__loadHistory:
            avgLoad += f[0]
        return avgLoad / len(self.__loadHistory), len(self.__loadHistory) == numRequiredSamples

    def __getNumJobWrappers(self):
        if not os.path.isdir(self.vmJobWrappersLocation):
            return 0
        self.log.info("VM job wrappers path: %s" % self.vmJobWrappersLocation)
        jdlList = glob.glob(os.path.join(self.vmJobWrappersLocation, "*.jdl"))
        jdlList += glob.glob(os.path.join(self.vmJobWrappersLocation, "*", "*.jdl"))
        return len(jdlList)

    def execute(self):

        # Get load
        avgLoad, avgRequiredSamples = self.__getLoadAvg()
        self.log.info("Load Average is %.2f" % avgLoad)
        if not avgRequiredSamples:
            self.log.info(" Not all required samples yet there")
        # Do we need to send heartbeat?
        with open("/proc/uptime") as fd:
            uptime = float(List.fromChar(fd.read().strip(), " ")[0])
        hours = int(uptime / 3600)
        minutes = int(uptime - hours * 3600) / 60
        seconds = uptime - hours * 3600 - minutes * 60
        self.log.info("Uptime is %.2f (%d:%02d:%02d)" % (uptime, hours, minutes, seconds))
        # Num jobs
        numJobs = self.__getNumJobWrappers()
        self.log.info("There are %d job wrappers" % numJobs)
        if uptime % self.heartBeatPeriod <= self.am_getPollingTime():
            # Heartbeat time!
            self.log.info("Sending hearbeat...")
            result = virtualMachineDB.instanceIDHeartBeat(
                self.uniqueID,
                avgLoad,
                numJobs,
                0,
                0,
            )
            status = None
            if result["OK"]:
                self.log.info(" heartbeat sent!")
                status = result["Value"]
            else:
                if "Transition" in result["Message"]:
                    self.log.error("Error on service:", result["Message"])
                    status = result["State"]
                else:
                    self.log.error("Connection error", result["Message"])
            if status:
                self.__processHeartBeatMessage(status, avgLoad)

        # Halt only if JobAgent is not running any longer
        jobAgentRunning = self.__isJobAgentRunning()
        self.log.info("JobAgent running: %s" % jobAgentRunning)
        if uptime % self.haltPeriod + self.haltBeforeMargin > self.haltPeriod:
            if not jobAgentRunning:
                self.log.info("JobAgent not running, stopping instance")
                self.__haltInstance(avgLoad)
        return S_OK()

    def __processHeartBeatMessage(self, hbMsg, avgLoad=0.0):
        if hbMsg == "stop":
            # Write stop file for jobAgent
            self.log.info("Received STOP signal. Writing stop files...")
            for agentName in ["WorkloadManagement/JobAgent"]:
                ad = os.path.join(*agentName.split("/"))
                stopDir = os.path.join(gConfig.getValue("/LocalSite/InstancePath", rootPath), "control", ad)
                stopFile = os.path.join(stopDir, "stop_agent")
                try:
                    if not os.path.isdir(stopDir):
                        os.makedirs(stopDir)
                    fd = open(stopFile, "w")
                    fd.write("stop!")
                    fd.close()
                    self.log.info("Wrote stop file %s for agent %s" % (stopFile, agentName))
                except Exception as e:
                    self.log.error("Could not write stop agent file", stopFile)
        if hbMsg == "halt":
            self.__haltInstance(avgLoad)

    def __haltInstance(self, avgLoad=0.0):
        self.log.info("Halting instance...")
        retries = 3
        sleepTime = 10
        for i in range(retries):
            result = virtualMachineDB.declareInstanceHalting(self.uniqueID, avgLoad)
            if result["OK"]:
                self.log.info("Declared instance halting")
                break
            self.log.error("Could not send halting state:", result["Message"])
            if i < retries - 1:
                self.log.info("Sleeping for %d seconds and retrying" % sleepTime)
                time.sleep(sleepTime)

        # self.log.info( "Executing system halt..." )
        # os.system( "halt" )



""" Main class for doing consistency checks, between files in:
    - File Catalog
    - TransformationSystem

    Should be extended to include the Storage (in DIRAC)
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import os
import time
import sys
import re
import errno

import DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.DataManagementSystem.Client.DataIntegrityClient import DataIntegrityClient
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.Core.Utilities.Adler import compareAdler


class ConsistencyInspector(object):
    """A class for handling some consistency checks"""

    def __init__(self, interactive=True, transClient=None, dm=None, fc=None, dic=None):
        """c'tor
        interactive: Data Manager (True) or DIRAC Agente (False)
        transClient: TransformationClient() if None, else transClient params
        dm: DataManager() if None, else dm params
        fc: FileCatalog() if None, else fc params
        One object for every production/directoriesList...
        """
        self.interactive = interactive
        self.transClient = TransformationClient() if transClient is None else transClient
        self.dataManager = dm if dm else DataManager()
        self.fileCatalog = fc if fc else FileCatalog()
        self.dic = dic if dic else DataIntegrityClient()
        self.dirac = Dirac()

        # Base elements from which to start the consistency checks
        self._prod = 0
        self._bkQuery = None
        self._fileType = []
        self._fileTypesExcluded = []
        self._lfns = []
        self.directories = []

        # Accessory elements
        self.runsList = []
        self.runStatus = None
        self.fromProd = None
        self.transType = ""
        self.cachedReplicas = {}

        self.prcdWithDesc = []
        self.prcdWithoutDesc = []
        self.prcdWithMultDesc = []
        self.nonPrcdWithDesc = []
        self.nonPrcdWithoutDesc = []
        self.nonPrcdWithMultDesc = []
        self.descForPrcdLFNs = []
        self.descForNonPrcdLFNs = []
        self.removedFiles = []

        self.absentLFNsInFC = []
        self.existLFNsNoSE = {}
        self.existLFNsBadReplicas = {}
        self.existLFNsBadFiles = {}
        self.existLFNsNotExisting = {}
        self.commonAncestors = {}
        self.multipleDescendants = {}
        self.ancestors = {}

        self._verbose = False

    def __logVerbose(self, msg, msg1=""):
        """logger helper for verbose information"""
        if self._verbose:
            newMsg = "[ConsistencyChecks] " + ("[%s] " % str(self.prod)) if self.prod else ""
            # Add that prefix to all lines of the message
            newMsg1 = msg1.replace("\n", "\n" + newMsg)
            newMsg += msg.replace("\n", "\n" + newMsg)
            gLogger.notice(newMsg, newMsg1)
        else:
            gLogger.verbose(msg, msg1)

    ##########################################################################

    def checkFC2SE(self):
        """check files vs SE information"""
        repDict = self.compareChecksum(self.lfns)
        self.existLFNsNoSE = repDict["MissingReplica"]
        self.existLFNsNotExisting = repDict["MissingAllReplicas"]
        self.existLFNsBadReplicas = repDict["SomeReplicasCorrupted"]
        self.existLFNsBadFiles = repDict["AllReplicasCorrupted"]

    def getReplicasPresence(self, lfns):
        """get the replicas using the standard FileCatalog.getReplicas()"""
        present = set()
        notPresent = set()

        chunkSize = 100
        printProgress = len(lfns) > chunkSize
        startTime = time.time()
        self.__write(
            "Checking replicas for %d files%s"
            % (len(lfns), (" (chunks of %d)" % chunkSize) if printProgress else "... ")
        )
        for chunk in breakListIntoChunks(lfns, chunkSize):
            if printProgress:
                self.__write(".")
            for _ in range(1, 10):
                res = self.fileCatalog.getReplicas(chunk)
                if res["OK"]:
                    present.update(res["Value"]["Successful"])
                    self.cachedReplicas.update(res["Value"]["Successful"])
                    notPresent.update(res["Value"]["Failed"])
                    break
                else:
                    time.sleep(0.1)
        self.__write(" (%.1f seconds)\n" % (time.time() - startTime))

        if notPresent:
            self.__logVerbose("Files without replicas:", "\n".join([""] + sorted(notPresent)))
        return list(present), list(notPresent)

    ##########################################################################

    def getReplicasPresenceFromDirectoryScan(self, lfns):
        """Get replicas scanning the directories. Might be faster."""

        dirs = {}
        present = []
        notPresent = []
        compare = True

        for lfn in lfns:
            dirN = os.path.dirname(lfn)
            if lfn == dirN + "/":
                compare = False
            dirs.setdefault(dirN, []).append(lfn)

        if compare:
            self.__write("Checking File Catalog for %d files from %d directories " % (len(lfns), len(dirs)))
        else:
            self.__write("Getting files from %d directories " % len(dirs))
        startTime = time.time()

        for dirN in sorted(dirs):
            startTime1 = time.time()
            self.__write(".")
            lfnsFound = self._getFilesFromDirectoryScan(dirN)
            gLogger.verbose("Obtained %d files in %.1f seconds" % (len(lfnsFound), time.time() - startTime1))
            if compare:
                pr, notPr = self.__compareLFNLists(dirs[dirN], lfnsFound)
                notPresent += notPr
                present += pr
            else:
                present += lfnsFound

        self.__write(" (%.1f seconds)\n" % (time.time() - startTime))
        gLogger.info("Found %d files with replicas and %d without" % (len(present), len(notPresent)))
        return present, notPresent

    ##########################################################################

    def __compareLFNLists(self, lfns, lfnsFound):
        """return files in both lists and files in lfns and not in lfnsFound"""
        present = []
        notPresent = lfns
        startTime = time.time()
        self.__logVerbose("Comparing list of %d LFNs with second list of %d" % (len(lfns), len(lfnsFound)))
        if lfnsFound:
            setLfns = set(lfns)
            setLfnsFound = set(lfnsFound)
            present = list(setLfns & setLfnsFound)
            notPresent = list(setLfns - setLfnsFound)
        self.__logVerbose("End of comparison: %.1f seconds" % (time.time() - startTime))
        return present, notPresent

    def _getFilesFromDirectoryScan(self, dirs):
        """calls dm.getFilesFromDirectory"""

        level = gLogger.getLevel()
        gLogger.setLevel("FATAL")
        res = self.dataManager.getFilesFromDirectory(dirs)
        gLogger.setLevel(level)
        if not res["OK"]:
            if "No such file or directory" not in res["Message"]:
                gLogger.error("Error getting files from directories %s:" % dirs, res["Message"])
            return []
        if res["Value"]:
            lfnsFound = res["Value"]
        else:
            lfnsFound = []

        return lfnsFound

    ##########################################################################

    def _getTSFiles(self):
        """Helper function - get files from the TS"""

        selectDict = {"TransformationID": self.prod}
        if self._lfns:
            selectDict["LFN"] = self._lfns
        elif self.runStatus and self.fromProd:
            res = self.transClient.getTransformationRuns({"TransformationID": self.fromProd, "Status": self.runStatus})
            if not res["OK"]:
                gLogger.error("Failed to get runs for transformation %d" % self.prod)
            else:
                if res["Value"]:
                    self.runsList.extend(
                        [run["RunNumber"] for run in res["Value"] if run["RunNumber"] not in self.runsList]
                    )
                    gLogger.notice("%d runs selected" % len(res["Value"]))
                elif not self.runsList:
                    gLogger.notice("No runs selected, check completed")
                    DIRAC.exit(0)
        if not self._lfns and self.runsList:
            selectDict["RunNumber"] = self.runsList

        res = self.transClient.getTransformation(self.prod)
        if not res["OK"]:
            gLogger.error("Failed to find transformation %s" % self.prod)
            return [], [], []
        status = res["Value"]["Status"]
        if status not in ("Active", "Stopped", "Completed", "Idle"):
            gLogger.notice(
                "Transformation %s in status %s, will not check if files are processed" % (self.prod, status)
            )
            processedLFNs = []
            nonProcessedLFNs = []
            nonProcessedStatuses = []
            if self._lfns:
                processedLFNs = self._lfns
        else:
            res = self.transClient.getTransformationFiles(selectDict)
            if not res["OK"]:
                gLogger.error("Failed to get files for transformation %d" % self.prod, res["Message"])
                return [], [], []
            else:
                processedLFNs = [item["LFN"] for item in res["Value"] if item["Status"] == "Processed"]
                nonProcessedLFNs = [item["LFN"] for item in res["Value"] if item["Status"] != "Processed"]
                nonProcessedStatuses = list(
                    set(item["Status"] for item in res["Value"] if item["Status"] != "Processed")
                )

        return processedLFNs, nonProcessedLFNs, nonProcessedStatuses

    def __getDirectories(self):
        """get the directories where to look into (they are either given, or taken from the transformation ID"""
        if self.directories:
            directories = []
            printout = False
            for directory in self.directories:
                if not directory.endswith("..."):
                    directories.append(directory)
                else:
                    printout = True
                    topDir = os.path.dirname(directory)
                    res = self.fileCatalog.listDirectory(topDir)
                    if not res["OK"]:
                        # DError(errno.ENOENT, res['Message'] )
                        return S_ERROR(errno.ENOENT, res["Message"])
                    else:
                        matchDir = directory.split("...")[0]
                        directories += [
                            d
                            for d in res["Value"]["Successful"].get(topDir, {}).get("SubDirs", [])
                            if d.startswith(matchDir)
                        ]
            if printout:
                gLogger.always("Expanded list of %d directories:\n%s" % (len(directories), "\n".join(directories)))
            return directories
        else:
            return S_ERROR(errno.ENOENT, "Need to specify the directories")

    ##########################################################################

    def __write(self, text):
        if self.interactive:
            sys.stdout.write(text)
            sys.stdout.flush()

    ##########################################################################

    def _selectByFileType(self, lfnDict, fileTypes=None, fileTypesExcluded=None):
        """Select only those files from the values of lfnDict that have a certain type"""
        if not lfnDict:
            return {}
        if not fileTypes:
            fileTypes = self.fileType
        if not fileTypesExcluded:
            fileTypesExcluded = self.fileTypesExcluded
        else:
            fileTypesExcluded += [ft for ft in self.fileTypesExcluded if ft not in fileTypesExcluded]
        # lfnDict is a dictionary of dictionaries including the metadata, create a
        # deep copy to get modified
        ancDict = dict(lfnDict)
        if fileTypes == [""]:
            fileTypes = []
        # and loop on the original dictionaries
        for ancestor in lfnDict:
            for desc in list(lfnDict[ancestor]):
                ft = lfnDict[ancestor][desc]["FileType"]
                if ft in fileTypesExcluded or (fileTypes and ft not in fileTypes):
                    ancDict[ancestor].pop(desc)
            if not len(ancDict[ancestor]):
                ancDict.pop(ancestor)
        return ancDict

    @staticmethod
    def _getFileTypesCount(lfnDict):
        """return file types count"""
        ft_dict = {}
        for ancestor in lfnDict:
            t_dict = {}
            for desc in lfnDict[ancestor]:
                ft = lfnDict[ancestor][desc]["FileType"]
                t_dict[ft] = t_dict.setdefault(ft, 0) + 1
            ft_dict[ancestor] = t_dict

        return ft_dict

    def __getLFNsFromFC(self):
        """Check if a list of LFNs is in the FC or not"""
        if not self.lfns:
            directories = []
            for dirName in self.__getDirectories():
                if not dirName.endswith("/"):
                    dirName += "/"
                directories.append(dirName)
            present, notPresent = self.getReplicasPresenceFromDirectoryScan(directories)
        else:
            present, notPresent = self.getReplicasPresence(self.lfns)
        return present, notPresent

    def compareChecksum(self, lfns):
        """compare the checksum of the file in the FC and the checksum of the physical replicas.
        Returns a dictionary containing 3 sub-dictionaries: one with files with missing PFN, one with
        files with all replicas corrupted, and one with files with some replicas corrupted and at least
        one good replica
        """
        retDict = {
            "AllReplicasCorrupted": {},
            "SomeReplicasCorrupted": {},
            "MissingReplica": {},
            "MissingAllReplicas": {},
            "NoReplicas": {},
        }

        chunkSize = 100
        replicas = {}
        setLfns = set(lfns)
        cachedLfns = setLfns & set(self.cachedReplicas)
        for lfn in cachedLfns:
            replicas[lfn] = self.cachedReplicas[lfn]
        lfnsLeft = list(setLfns - cachedLfns)
        if lfnsLeft:
            self.__write("Get replicas for %d files (chunks of %d): " % (len(lfnsLeft), chunkSize))
            for lfnChunk in breakListIntoChunks(lfnsLeft, chunkSize):
                self.__write(".")
                replicasRes = self.fileCatalog.getReplicas(lfnChunk)
                if not replicasRes["OK"]:
                    gLogger.error("error:  %s" % replicasRes["Message"])
                    return S_ERROR(errno.ENOENT, "error:  %s" % replicasRes["Message"])
                replicasRes = replicasRes["Value"]
                if replicasRes["Failed"]:
                    retDict["NoReplicas"].update(replicasRes["Failed"])
                replicas.update(replicasRes["Successful"])

        self.__write("Get FC metadata for %d files to be checked: " % len(lfns))
        metadata = {}
        for lfnChunk in breakListIntoChunks(replicas, chunkSize):
            self.__write(".")
            res = self.fileCatalog.getFileMetadata(lfnChunk)
            if not res["OK"]:
                return S_ERROR(errno.ENOENT, "error %s" % res["Message"])
            metadata.update(res["Value"]["Successful"])

        gLogger.notice("Check existence and compare checksum file by file...")
        csDict = {}
        seFiles = {}
        # Reverse the LFN->SE dictionary
        nReps = 0
        for lfn in replicas:
            csDict.setdefault(lfn, {})["FCChecksum"] = metadata.get(lfn, {}).get("Checksum")
            for se in replicas[lfn]:
                seFiles.setdefault(se, []).append(lfn)
                nReps += 1

        gLogger.notice("Getting checksum of %d replicas in %d SEs" % (nReps, len(seFiles)))
        checkSum = {}
        lfnNotExisting = {}
        lfnNoInfo = {}
        logLevel = gLogger.getLevel()
        gLogger.setLevel("FATAL")
        for num, se in enumerate(sorted(seFiles)):
            self.__write("\n%d. At %s (%d files): " % (num, se, len(seFiles[se])))
            oSe = StorageElement(se)
            notFound = 0
            for surlChunk in breakListIntoChunks(seFiles[se], chunkSize):
                self.__write(".")
                metadata = oSe.getFileMetadata(surlChunk)
                if not metadata["OK"]:
                    gLogger.error("Error: getFileMetadata returns %s. Ignore those replicas" % (metadata["Message"]))
                    # Remove from list of replicas as we don't know whether it is OK or
                    # not
                    for lfn in seFiles[se]:
                        lfnNoInfo.setdefault(lfn, []).append(se)
                else:
                    metadata = metadata["Value"]
                    notFound += len(metadata["Failed"])
                    for lfn in metadata["Failed"]:
                        lfnNotExisting.setdefault(lfn, []).append(se)
                    for lfn in metadata["Successful"]:
                        checkSum.setdefault(lfn, {})[se] = metadata["Successful"][lfn]["Checksum"]
            if notFound:
                gLogger.error("%d files not found" % notFound)

        gLogger.setLevel(logLevel)

        gLogger.notice("Verifying checksum of %d files" % len(replicas))
        for lfn in replicas:
            # get the lfn checksum from the FC
            replicaDict = replicas[lfn]
            oneGoodReplica = False
            allGoodReplicas = True
            fcChecksum = csDict[lfn].pop("FCChecksum")
            for se in replicaDict:
                # If replica doesn't exist skip check
                if se in lfnNotExisting.get(lfn, []):
                    allGoodReplicas = False
                    continue
                if se in lfnNoInfo.get(lfn, []):
                    # If there is no info, a priori it could be good
                    oneGoodReplica = True
                    continue
                # get the surls metadata and compare the checksum
                surlChecksum = checkSum.get(lfn, {}).get(se, "")
                if not surlChecksum or not compareAdler(fcChecksum, surlChecksum):
                    # if fcChecksum does not match surlChecksum
                    csDict[lfn][se] = {"PFNChecksum": surlChecksum}
                    gLogger.info(
                        "ERROR!! checksum mismatch at %s for LFN %s:  FC checksum: %s , PFN checksum : %s "
                        % (se, lfn, fcChecksum, surlChecksum)
                    )
                    allGoodReplicas = False
                else:
                    oneGoodReplica = True
            if not oneGoodReplica:
                if lfn in lfnNotExisting:
                    gLogger.info("=> All replicas are missing", lfn)
                    retDict["MissingAllReplicas"][lfn] = "All"
                else:
                    gLogger.info("=> All replicas have bad checksum", lfn)
                    retDict["AllReplicasCorrupted"][lfn] = csDict[lfn]
            elif not allGoodReplicas:
                if lfn in lfnNotExisting:
                    gLogger.info("=> At least one replica missing", lfn)
                    retDict["MissingReplica"][lfn] = lfnNotExisting[lfn]
                else:
                    gLogger.info("=> At least one replica with good Checksum", lfn)
                    retDict["SomeReplicasCorrupted"][lfn] = csDict[lfn]

        return S_OK(retDict)

    ##########################################################################
    # properties

    def set_prod(self, value):
        """Setter"""
        if value:
            value = int(value)
            res = self.transClient.getTransformation(value, extraParams=False)
            if not res["OK"]:
                raise Exception("Couldn't find transformation %d: %s" % (value, res["Message"]))
            else:
                self.transType = res["Value"]["Type"]
            if self.interactive:
                gLogger.info("Production %d has type %s" % (value, self.transType))
        else:
            value = 0
        self._prod = value

    def get_prod(self):
        """Getter"""
        return self._prod

    prod = property(get_prod, set_prod)

    def set_fileType(self, value):
        """Setter"""
        self._fileType = [ft.upper() for ft in value]

    def get_fileType(self):
        """Getter"""
        return self._fileType

    fileType = property(get_fileType, set_fileType)

    def set_fileTypesExcluded(self, value):
        """Setter"""
        self._fileTypesExcluded = [ft.upper() for ft in value]

    def get_fileTypesExcluded(self):
        """Getter"""
        return self._fileTypesExcluded

    fileTypesExcluded = property(get_fileTypesExcluded, set_fileTypesExcluded)

    def set_lfns(self, value):
        """Setter"""
        if isinstance(value, six.string_types):
            value = [value]
        value = [v.replace(" ", "").replace("//", "/") for v in value]
        self._lfns = value

    def get_lfns(self):
        """Getter"""
        return self._lfns

    lfns = property(get_lfns, set_lfns)

    ##########################################################################
    #
    #  This part was backported from DataIntegrityClient
    #
    #
    #  This section contains the specific methods for File Catalog->SE checks
    #

    def catalogDirectoryToSE(self, lfnDir):
        """This obtains the replica and metadata information from the catalog
        for the supplied directory and checks against the storage elements.
        """
        gLogger.info("-" * 40)
        gLogger.info("Performing the FC->SE check")
        gLogger.info("-" * 40)
        if isinstance(lfnDir, six.string_types):
            lfnDir = [lfnDir]
        res = self._getCatalogDirectoryContents(lfnDir)
        if not res["OK"]:
            return res
        replicas = res["Value"]["Replicas"]
        catalogMetadata = res["Value"]["Metadata"]
        res = self.checkPhysicalFiles(replicas, catalogMetadata)
        if not res["OK"]:
            return res
        resDict = {"CatalogMetadata": catalogMetadata, "CatalogReplicas": replicas}
        return S_OK(resDict)

    def catalogFileToSE(self, lfns):
        """This obtains the replica and metadata information from the catalog and checks against the storage elements."""
        gLogger.info("-" * 40)
        gLogger.info("Performing the FC->SE check")
        gLogger.info("-" * 40)
        if isinstance(lfns, six.string_types):
            lfns = [lfns]
        res = self._getCatalogMetadata(lfns)
        if not res["OK"]:
            return res
        catalogMetadata, _missingCatalogFiles, _zeroSizeFiles = res["Value"]
        res = self._getCatalogReplicas(list(catalogMetadata))
        if not res["OK"]:
            return res
        replicas, _zeroReplicaFiles = res["Value"]
        res = self.checkPhysicalFiles(replicas, catalogMetadata)
        if not res["OK"]:
            return res
        resDict = {"CatalogMetadata": catalogMetadata, "CatalogReplicas": replicas}
        return S_OK(resDict)

    def checkPhysicalFiles(self, replicas, catalogMetadata, ses=None):
        """This method takes the supplied replica and metadata information obtained
        from the catalog and checks against the storage elements.
        """

        # FIXME: we better use the compareChecksum function instead of this one!
        # or maybe directly checkFC2SE

        gLogger.info("-" * 40)
        gLogger.info("Performing the FC->SE check")
        gLogger.info("-" * 40)
        seLfns = {}
        for lfn, replicaDict in replicas.items():
            for se, _url in replicaDict.items():
                if (ses) and (se not in ses):
                    continue
                seLfns.setdefault(se, []).append(lfn)
        gLogger.info("%s %s" % ("Storage Element".ljust(20), "Replicas".rjust(20)))

        for se in sorted(seLfns):
            files = len(seLfns[se])
            gLogger.info("%s %s" % (se.ljust(20), str(files).rjust(20)))

            lfns = seLfns[se]
            sizeMismatch = []
            res = self.__checkPhysicalFileMetadata(lfns, se)
            if not res["OK"]:
                gLogger.error("Failed to get physical file metadata.", res["Message"])
                return res
            for lfn, metadata in res["Value"].items():
                if lfn in catalogMetadata:
                    # and ( metadata['Size'] != 0 ):
                    if metadata["Size"] != catalogMetadata[lfn]["Size"]:
                        sizeMismatch.append((lfn, "deprecatedUrl", se, "CatalogPFNSizeMismatch"))
            if sizeMismatch:
                self.dic.reportProblematicReplicas(sizeMismatch, se, "CatalogPFNSizeMismatch")
        return S_OK()

    def __checkPhysicalFileMetadata(self, lfns, se):
        """Check obtain the physical file metadata and check the files are available"""
        gLogger.info("Checking the integrity of %s physical files at %s" % (len(lfns), se))

        res = StorageElement(se).getFileMetadata(lfns)

        if not res["OK"]:
            gLogger.error("Failed to get metadata for lfns.", res["Message"])
            return res
        pfnMetadata = res["Value"]["Successful"]
        # If the replicas are completely missing
        missingReplicas = []
        for lfn, reason in res["Value"]["Failed"].items():
            if re.search("File does not exist", reason):
                missingReplicas.append((lfn, "deprecatedUrl", se, "PFNMissing"))
        if missingReplicas:
            self.dic.reportProblematicReplicas(missingReplicas, se, "PFNMissing")
        lostReplicas = []
        unavailableReplicas = []
        zeroSizeReplicas = []
        # If the files are not accessible
        for lfn, metadata in pfnMetadata.items():
            if metadata.get("Lost"):
                lostReplicas.append((lfn, "deprecatedUrl", se, "PFNLost"))
            if metadata.get("Unavailable") or not metadata["Accessible"]:
                unavailableReplicas.append((lfn, "deprecatedUrl", se, "PFNUnavailable"))
            if not metadata["Size"]:
                zeroSizeReplicas.append((lfn, "deprecatedUrl", se, "PFNZeroSize"))
        if lostReplicas:
            self.dic.reportProblematicReplicas(lostReplicas, se, "PFNLost")
        if unavailableReplicas:
            self.dic.reportProblematicReplicas(unavailableReplicas, se, "PFNUnavailable")
        if zeroSizeReplicas:
            self.dic.reportProblematicReplicas(zeroSizeReplicas, se, "PFNZeroSize")
        gLogger.info("Checking the integrity of physical files at %s complete" % se)
        return S_OK(pfnMetadata)

    ##########################################################################
    #
    # This section contains the specific methods for SE->File Catalog checks
    #

    def _getCatalogDirectoryContents(self, lfnDirs):
        """Obtain the contents of the supplied directory, recursively"""

        def _getDirectoryContent(directory):
            """Inner function: recursively scan a directory, returns list of LFNs"""
            filesInDirectory = {}

            gLogger.debug("Examining %s" % directory)

            res = self.fileCatalog.listDirectory(directory)
            if not res["OK"]:
                gLogger.error("Failed to get directory contents", res["Message"])
                return res
            if directory in res["Value"]["Failed"]:
                gLogger.error(
                    "Failed to get directory content", "%s %s" % (directory, res["Value"]["Failed"][directory])
                )
                return S_ERROR("Failed to get directory content")
            if directory not in res["Value"]["Successful"]:
                return S_ERROR("Directory not existing?")

            # first, adding the files found in the current directory
            gLogger.debug("Files in %s: %d" % (directory, len(res["Value"]["Successful"][directory]["Files"])))
            filesInDirectory.update(res["Value"]["Successful"][directory]["Files"])

            # then, looking for subDirectories content
            if res["Value"]["Successful"][directory]["SubDirs"]:
                for l_dir in res["Value"]["Successful"][directory]["SubDirs"]:
                    # recursion here
                    subDirContent = _getDirectoryContent(l_dir)
                    if not subDirContent["OK"]:
                        return subDirContent
                    else:
                        filesInDirectory.update(subDirContent["Value"])

            return S_OK(filesInDirectory)

        gLogger.info("Obtaining the catalog contents for %d directories" % len(lfnDirs))

        allFiles = {}
        for lfnDir in lfnDirs:
            dirContent = _getDirectoryContent(lfnDir)
            if not dirContent["OK"]:
                return dirContent
            else:
                gLogger.debug("Content of directory %s: %d files" % (lfnDir, len(dirContent["Value"])))
                allFiles.update(dirContent["Value"])

        gLogger.debug("Content of directories examined: %d files" % len(allFiles))

        replicas = self.fileCatalog.getReplicas(list(allFiles))
        if not replicas["OK"]:
            return replicas
        if replicas["Value"]["Failed"]:
            return S_ERROR("Failures in replicas discovery")

        return S_OK({"Metadata": allFiles, "Replicas": replicas["Value"]["Successful"]})

    def _getCatalogReplicas(self, lfns):
        """Obtain the file replicas from the catalog while checking that there are replicas"""
        if not lfns:
            return S_OK(([], []))

        gLogger.info("Obtaining the replicas for %s files" % len(lfns))
        zeroReplicaFiles = []
        res = self.fileCatalog.getReplicas(lfns, allStatus=True)
        if not res["OK"]:
            gLogger.error("Failed to get catalog replicas", res["Message"])
            return res
        allReplicas = res["Value"]["Successful"]
        for lfn, error in res["Value"]["Failed"].items():
            if re.search("File has zero replicas", error):
                zeroReplicaFiles.append(lfn)
        gLogger.info("Obtaining the replicas for files complete")
        return S_OK((allReplicas, zeroReplicaFiles))

    def _getCatalogMetadata(self, lfns):
        """Obtain the file metadata from the catalog while checking they exist"""
        allMetadata = []
        missingCatalogFiles = []
        zeroSizeFiles = []

        if not lfns:
            return S_OK((allMetadata, missingCatalogFiles, zeroSizeFiles))
        gLogger.info("Obtaining the catalog metadata for %s files" % len(lfns))

        res = self.fileCatalog.getFileMetadata(lfns)
        if not res["OK"]:
            gLogger.error("Failed to get catalog metadata", res["Message"])
            return res
        allMetadata = res["Value"]["Successful"]
        for lfn, error in res["Value"]["Failed"].items():
            if re.search("No such file or directory", error):
                missingCatalogFiles.append(lfn)
        gLogger.info("Obtaining the catalog metadata complete")
        return S_OK((allMetadata, missingCatalogFiles, zeroSizeFiles))



""" DIRAC FileCatalog component representing a simple directory tree
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id:  $"

import os

from DIRAC import S_OK, S_ERROR
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.DirectoryManager.DirectoryTreeBase import DirectoryTreeBase


class DirectorySimpleTree(DirectoryTreeBase):
    """Class managing Directory Tree as a simple self-linked structure with full
    directory path stored in each node
    """

    def __init__(self, database=None):
        DirectoryTreeBase.__init__(self, database)
        self.treeTable = "FC_DirectoryTree"

    def findDir(self, path):

        req = "SELECT DirID from FC_DirectoryTree WHERE DirName='%s'" % path
        result = self.db._query(req)
        if not result["OK"]:
            return result

        if not result["Value"]:
            return S_OK("")

        return S_OK(result["Value"][0][0])

    def removeDir(self, path):
        """Remove directory"""

        result = self.findDir(path)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK()

        dirID = result["Value"]
        req = "DELETE FROM FC_DirectoryTree WHERE DirID=%d" % dirID
        result = self.db._update(req)
        return result

    def makeDir(self, path):

        result = self.findDir(path)
        if not result["OK"]:
            return result
        dirID = result["Value"]
        if dirID:
            return S_OK(dirID)
        names = ["DirName"]
        values = [path]
        result = self.db.insertFields("FC_DirectoryTree", names, values)
        if not result["OK"]:
            return result
        return S_OK(result["lastRowId"])

    def existsDir(self, path):
        """Check the existence of a directory at the specified path"""
        result = self.findDir(path)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK({"Exists": False})
        else:
            return S_OK({"Exists": True, "DirID": result["Value"]})

    def getParent(self, path):
        """Get the parent ID of the given directory"""
        parent_dir = os.path.dirname(path)
        if parent_dir == "/":
            return S_OK(0)
        return self.findDir(parent_dir)

    def getParentID(self, dirID):
        """Get the ID of the parent of a directory specified by ID"""
        if dirID == 0:
            return S_ERROR("Root directory ID given")

        req = "SELECT Parent FROM FC_DirectoryTree WHERE DirID=%d" % dirID
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("No parent found")

        return S_OK(result["Value"][0][0])

    def getDirectoryPath(self, dirID):
        """Get directory name by directory ID"""
        req = "SELECT DirName FROM FC_DirectoryTree WHERE DirID=%d" % int(dirID)
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Directory with id %d not found" % int(dirID))

        return S_OK(result["Value"][0][0])

    def getDirectoryName(self, dirID):
        """Get directory name by directory ID"""

        result = self.getDirectoryPath(dirID)
        if not result["OK"]:
            return result

        return S_OK(os.path.basename(result["Value"]))

    def getPathIDs(self, path):
        """Get IDs of all the directories in the parent hierarchy"""

        elements = path.split("/")
        pelements = []
        dPath = ""
        for el in elements[1:]:
            dPath += "/" + el
            pelements.append(dPath)

        pathString = ["'" + p + "'" for p in pelements]
        req = "SELECT DirID FROM FC_DirectoryTree WHERE DirName in (%s) ORDER BY DirID" % pathString
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Directory %s not found" % path)

        return S_OK([x[0] for x in result["Value"]])

    def getChildren(self, path):
        """Get child directory IDs for the given directory"""
        if isinstance(path, str):
            result = self.findDir(path)
            if not result["OK"]:
                return result
            dirID = result["Value"]
        else:
            dirID = path

        req = "SELECD DirID FROM FC_DirectoryTree WHERE Parent=%d" % dirID
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK([])

        return S_OK([x[0] for x in result["Value"]])



"""
Utilities to execute one or more functions with a given proxy.

:func:`executeWithUserProxy` decorator example usage::

  @executeWithUserProxy
  def testFcn(x, i, kw='qwerty'):

    print "args", x, i
    print "kwargs", kw
    print os.environ.get('X509_USER_PROXY')
    return S_OK()

  ...

  result = testFcn( 1.0, 1, kw = 'asdfghj', proxyUserName = 'atsareg', proxyUserGroup = 'biomed_user' )


:func:`executeWithUserProxy` function wrapper example::

  def undecoratedFunction(foo='bar'):
     print foo, os.environ.get('X509_USER_PROXY')
     return S_OK()

  executeWithUserProxy(testFcn)(foo='baz', proxyUserName='atsareg', proxyUserGroup='biomed_user')


:class:`UserProxy` context manager example::

  with UserProxy(proxyUserName='user', proxyUserGroup='group') as proxyResult:
    if proxyResult['OK']:
      functionThatNeedsAProxy()
      anotherFunction()


"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from DIRAC import gConfig, gLogger, S_ERROR, S_OK
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOMSAttributeForGroup, getDNForUsername
from DIRAC.Core.Utilities.LockRing import LockRing

__RCSID__ = "$Id$"


def executeWithUserProxy(fcn):
    """
    Decorator function to execute with a temporary user proxy

    :param fcn: function to be decorated
    :return: the result of the fcn execution

    In order to be executed with a user proxy, the function must be called with the
    following parameters:

    :param str proxyUserName: the user name of the proxy to be used
    :param str proxyUserGroup: the user group of the proxy to be used
    :param str proxyUserDN: the user DN of the proxy to be used
    :param str proxyWithVOMS: optional flag to dress or not the user proxy with VOMS extension ( default True )
    :param str proxyFilePath: optional file location for the temporary proxy
    :param bool executionLock: flag to execute with a lock for the time of user proxy application ( default False )
    """

    def wrapped_fcn(*args, **kwargs):

        userName = kwargs.pop("proxyUserName", "")
        userDN = kwargs.pop("proxyUserDN", "")
        userGroup = kwargs.pop("proxyUserGroup", "")
        vomsFlag = kwargs.pop("proxyWithVOMS", True)
        proxyFilePath = kwargs.pop("proxyFilePath", False)
        executionLockFlag = kwargs.pop("executionLock", False)

        if (userName or userDN) and userGroup:

            proxyResults = _putProxy(
                userName=userName,
                userDN=userDN,
                userGroup=userGroup,
                vomsFlag=vomsFlag,
                proxyFilePath=proxyFilePath,
                executionLockFlag=executionLockFlag,
            )
            if not proxyResults["OK"]:
                return proxyResults
            originalUserProxy, useServerCertificate, executionLock = proxyResults["Value"]

            try:
                return fcn(*args, **kwargs)
            except Exception as lException:  # pylint: disable=broad-except
                value = ",".join([str(arg) for arg in lException.args])
                exceptType = lException.__class__.__name__
                return S_ERROR("Exception - %s: %s" % (exceptType, value))
            finally:
                _restoreProxyState(originalUserProxy, useServerCertificate, executionLock)
        else:
            # No proxy substitution requested
            return fcn(*args, **kwargs)

    return wrapped_fcn


def getProxy(userDNs, userGroup, vomsAttr, proxyFilePath):
    """do the actual download of the proxy, trying the different DNs"""
    for userDN in userDNs:
        if vomsAttr:
            result = gProxyManager.downloadVOMSProxyToFile(
                userDN,
                userGroup,
                requiredVOMSAttribute=vomsAttr,
                filePath=proxyFilePath,
                requiredTimeLeft=3600,
                cacheTime=3600,
            )
        else:
            result = gProxyManager.downloadProxyToFile(
                userDN, userGroup, filePath=proxyFilePath, requiredTimeLeft=3600, cacheTime=3600
            )

        if not result["OK"]:
            gLogger.error(
                "Can't download %sproxy " % ("VOMS" if vomsAttr else ""),
                "of '%s', group %s to file: " % (userDN, userGroup) + result["Message"],
            )
        else:
            return result

    # If proxy not found for any DN, return an error
    return S_ERROR("Can't download proxy")


def executeWithoutServerCertificate(fcn):
    """
    Decorator function to execute a call without the server certificate.
    This shows useful in Agents when we want to call a DIRAC service
    and use the shifter proxy (for example Write calls to the DFC).

    The method does not fetch any proxy, it assumes it is already
    set up in the environment.
    Note that because it modifies the configuration for all thread,
    it uses a lock (the same as ExecuteWithUserProxy)

    Potential problem:
      * there is a lock for this particular method, but any other method
        changing the UseServerCertificate value can clash with this.

    :param fcn: function to be decorated
    :return: the result of the fcn execution

    """

    def wrapped_fcn(*args, **kwargs):

        # Get the lock and acquire it
        executionLock = LockRing().getLock("_UseUserProxy_", recursive=True)
        executionLock.acquire()

        # Check if the caller is executing with the host certificate
        useServerCertificate = gConfig.useServerCertificate()
        if useServerCertificate:
            gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "false")

        try:
            return fcn(*args, **kwargs)
        except Exception as lException:  # pylint: disable=broad-except
            value = ",".join([str(arg) for arg in lException.args])
            exceptType = lException.__class__.__name__
            return S_ERROR("Exception - %s: %s" % (exceptType, value))
        finally:
            # Restore the default host certificate usage if necessary
            if useServerCertificate:
                gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "true")
            # release the lock
            executionLock.release()

    return wrapped_fcn


class UserProxy(object):
    """Implement a context manager to execute functions with a user proxy."""

    def __init__(
        self,
        proxyUserName=None,
        proxyUserGroup=None,
        proxyUserDN=None,
        proxyWithVOMS=True,
        proxyFilePath=None,
        executionLock=False,
    ):
        """Construct the context manager for a user proxy.

        :param str proxyUserName: the user name of the proxy to be used
        :param str proxyUserGroup: the user group of the proxy to be used
        :param str proxyUserDN: the user DN of the proxy to be used
        :param bool proxyWithVOMS: optional flag to dress or not the user proxy with VOMS extension ( default True )
        :param str proxyFilePath: optional file location for the temporary proxy
        :param bool executionLock: flag to execute with a lock for the time of user proxy application ( default False )
        """
        if not ((proxyUserName or proxyUserDN) and proxyUserGroup):
            self.result = S_OK()
        else:
            self.result = _putProxy(
                userDN=proxyUserDN,
                userName=proxyUserName,
                userGroup=proxyUserGroup,
                vomsFlag=proxyWithVOMS,
                executionLockFlag=executionLock,
                proxyFilePath=proxyFilePath,
            )

    def __enter__(self):
        return self.result

    def __exit__(self, _type, _value, _traceback):
        """Reset the proxy and certificate state at the end of the managed block."""
        if self.result["OK"] and self.result["Value"]:
            originalUserProxy, useServerCertificate, executionLock = self.result["Value"]
            _restoreProxyState(originalUserProxy, useServerCertificate, executionLock)


def _putProxy(userDN=None, userName=None, userGroup=None, vomsFlag=None, proxyFilePath=None, executionLockFlag=False):
    """Download proxy, place in a file and populate X509_USER_PROXY environment variable.

    Parameters like `userProxy` or `executeWithUserProxy`.
    :returns: Tuple of originalUserProxy, useServerCertificate, executionLock
    """
    # Setup user proxy
    if userDN:
        userDNs = [userDN]
    else:
        result = getDNForUsername(userName)
        if not result["OK"]:
            return result
        userDNs = result["Value"]  # a same user may have more than one DN

    vomsAttr = ""
    if vomsFlag:
        vomsAttr = getVOMSAttributeForGroup(userGroup)

    result = getProxy(userDNs, userGroup, vomsAttr, proxyFilePath)

    if not result["OK"]:
        return result

    executionLock = LockRing().getLock("_UseUserProxy_", recursive=True) if executionLockFlag else None
    if executionLockFlag:
        executionLock.acquire()

    os.environ["X509_USER_PROXY"], originalUserProxy = result["Value"], os.environ.get("X509_USER_PROXY")

    # Check if the caller is executing with the host certificate
    useServerCertificate = gConfig.useServerCertificate()
    if useServerCertificate:
        gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "false")

    return S_OK((originalUserProxy, useServerCertificate, executionLock))


def _restoreProxyState(originalUserProxy=None, useServerCertificate=None, executionLock=None):
    """Restore the default host certificate and proxy state if necessary.

    Parameters like the output from `_putProxy`.
    """
    if useServerCertificate:
        gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "true")
    if originalUserProxy:
        os.environ["X509_USER_PROXY"] = originalUserProxy
    else:
        os.environ.pop("X509_USER_PROXY", None)
    if executionLock:
        executionLock.release()



#!/usr/bin/env python
########################################################################
# File :    dirac-proxy-init.py
# Author :  Adrian Casajus
########################################################################
"""
Creating a proxy.

Example:
  $ dirac-proxy-init -g dirac_user -t --rfc
  Enter Certificate password:
"""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import glob
import time
import datetime

import DIRAC

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.FrameworkSystem.Client import ProxyGeneration, ProxyUpload
from DIRAC.Core.Security import X509Chain, ProxyInfo, VOMS
from DIRAC.Core.Security.Locations import getCAsLocation
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.FrameworkSystem.Client.BundleDeliveryClient import BundleDeliveryClient

__RCSID__ = "$Id$"


class Params(ProxyGeneration.CLIParams):

    addVOMSExt = False
    uploadProxy = True
    uploadPilot = False

    def setVOMSExt(self, _arg):
        self.addVOMSExt = True
        return S_OK()

    def disableProxyUpload(self, _arg):
        self.uploadProxy = False
        return S_OK()

    def registerCLISwitches(self):
        ProxyGeneration.CLIParams.registerCLISwitches(self)
        Script.registerSwitch(
            "U", "upload", "Upload a long lived proxy to the ProxyManager (deprecated, see --no-upload)"
        )
        Script.registerSwitch(
            "N", "no-upload", "Do not upload a long lived proxy to the ProxyManager", self.disableProxyUpload
        )
        Script.registerSwitch("M", "VOMS", "Add voms extension", self.setVOMSExt)


class ProxyInit(object):
    def __init__(self, piParams):
        self.__piParams = piParams
        self.__issuerCert = False
        self.__proxyGenerated = False
        self.__uploadedInfo = {}

    def getIssuerCert(self):
        if self.__issuerCert:
            return self.__issuerCert
        proxyChain = X509Chain.X509Chain()
        resultProxyChainFromFile = proxyChain.loadChainFromFile(self.__piParams.certLoc)
        if not resultProxyChainFromFile["OK"]:
            gLogger.error("Could not load the proxy: %s" % resultProxyChainFromFile["Message"])
            sys.exit(1)
        resultIssuerCert = proxyChain.getIssuerCert()
        if not resultIssuerCert["OK"]:
            gLogger.error("Could not load the proxy: %s" % resultIssuerCert["Message"])
            sys.exit(1)
        self.__issuerCert = resultIssuerCert["Value"]
        return self.__issuerCert

    def certLifeTimeCheck(self):
        minLife = Registry.getGroupOption(self.__piParams.diracGroup, "SafeCertificateLifeTime", 2592000)
        resultIssuerCert = self.getIssuerCert()
        resultRemainingSecs = resultIssuerCert.getRemainingSecs()  # pylint: disable=no-member
        if not resultRemainingSecs["OK"]:
            gLogger.error("Could not retrieve certificate expiration time", resultRemainingSecs["Message"])
            return
        lifeLeft = resultRemainingSecs["Value"]
        if minLife > lifeLeft:
            daysLeft = int(lifeLeft / 86400)
            msg = "Your certificate will expire in less than %d days. Please renew it!" % daysLeft
            sep = "=" * (len(msg) + 4)
            msg = "%s\n  %s  \n%s" % (sep, msg, sep)
            gLogger.notice(msg)

    def addVOMSExtIfNeeded(self):
        addVOMS = self.__piParams.addVOMSExt or Registry.getGroupOption(
            self.__piParams.diracGroup, "AutoAddVOMS", False
        )
        if not addVOMS:
            return S_OK()

        vomsAttr = Registry.getVOMSAttributeForGroup(self.__piParams.diracGroup)
        if not vomsAttr:
            return S_ERROR(
                "Requested adding a VOMS extension but no VOMS attribute defined for group %s"
                % self.__piParams.diracGroup
            )

        resultVomsAttributes = VOMS.VOMS().setVOMSAttributes(
            self.__proxyGenerated, attribute=vomsAttr, vo=Registry.getVOMSVOForGroup(self.__piParams.diracGroup)
        )
        if not resultVomsAttributes["OK"]:
            return S_ERROR(
                "Could not add VOMS extensions to the proxy\nFailed adding VOMS attribute: %s"
                % resultVomsAttributes["Message"]
            )

        gLogger.notice("Added VOMS attribute %s" % vomsAttr)
        chain = resultVomsAttributes["Value"]
        result = chain.dumpAllToFile(self.__proxyGenerated)
        if not result["OK"]:
            return result
        return S_OK()

    def createProxy(self):
        """Creates the proxy on disk"""
        gLogger.notice("Generating proxy...")
        resultProxyGenerated = ProxyGeneration.generateProxy(piParams)
        if not resultProxyGenerated["OK"]:
            gLogger.error(resultProxyGenerated["Message"])
            sys.exit(1)
        self.__proxyGenerated = resultProxyGenerated["Value"]
        return resultProxyGenerated

    def uploadProxy(self):
        """Upload the proxy to the proxyManager service"""
        issuerCert = self.getIssuerCert()
        resultUserDN = issuerCert.getSubjectDN()  # pylint: disable=no-member
        if not resultUserDN["OK"]:
            return resultUserDN
        userDN = resultUserDN["Value"]

        gLogger.notice("Uploading proxy..")
        if userDN in self.__uploadedInfo:
            expiry = self.__uploadedInfo[userDN].get("")
            if expiry:
                if (
                    issuerCert.getNotAfterDate()["Value"] - datetime.timedelta(minutes=10) < expiry
                ):  # pylint: disable=no-member
                    gLogger.info('Proxy with DN "%s" already uploaded' % userDN)
                    return S_OK()
        gLogger.info("Uploading %s proxy to ProxyManager..." % userDN)
        upParams = ProxyUpload.CLIParams()
        upParams.onTheFly = True
        upParams.proxyLifeTime = issuerCert.getRemainingSecs()["Value"] - 300  # pylint: disable=no-member
        upParams.rfcIfPossible = self.__piParams.rfc
        for k in ("certLoc", "keyLoc", "userPasswd"):
            setattr(upParams, k, getattr(self.__piParams, k))
        resultProxyUpload = ProxyUpload.uploadProxy(upParams)
        if not resultProxyUpload["OK"]:
            gLogger.error(resultProxyUpload["Message"])
            return resultProxyUpload
        self.__uploadedInfo = resultProxyUpload["Value"]
        gLogger.info("Proxy uploaded")
        return S_OK()

    def printInfo(self):
        """Printing utilities"""
        resultProxyInfoAsAString = ProxyInfo.getProxyInfoAsString(self.__proxyGenerated)
        if not resultProxyInfoAsAString["OK"]:
            gLogger.error("Failed to get the new proxy info: %s" % resultProxyInfoAsAString["Message"])
        else:
            gLogger.notice("Proxy generated:")
            gLogger.notice(resultProxyInfoAsAString["Value"])
        if self.__uploadedInfo:
            gLogger.notice("\nProxies uploaded:")
            maxDNLen = 0
            maxGroupLen = 0
            for userDN in self.__uploadedInfo:
                maxDNLen = max(maxDNLen, len(userDN))
                for group in self.__uploadedInfo[userDN]:
                    maxGroupLen = max(maxGroupLen, len(group))
            gLogger.notice(" %s | %s | Until (GMT)" % ("DN".ljust(maxDNLen), "Group".ljust(maxGroupLen)))
            for userDN in self.__uploadedInfo:
                for group in self.__uploadedInfo[userDN]:
                    gLogger.notice(
                        " %s | %s | %s"
                        % (
                            userDN.ljust(maxDNLen),
                            group.ljust(maxGroupLen),
                            self.__uploadedInfo[userDN][group].strftime("%Y/%m/%d %H:%M"),
                        )
                    )

    def checkCAs(self):
        caDir = getCAsLocation()
        if not caDir:
            gLogger.warn("No valid CA dir found.")
            return
        # In globus standards .r0 files are CRLs. They have the same names of the CAs but diffent file extension
        searchExp = os.path.join(caDir, "*.r0")
        crlList = glob.glob(searchExp)
        if not crlList:
            gLogger.warn("No CRL files found for %s. Abort check of CAs" % searchExp)
            return
        newestFPath = max(crlList, key=os.path.getmtime)
        newestFTime = os.path.getmtime(newestFPath)
        if newestFTime > (time.time() - (2 * 24 * 3600)):
            # At least one of the files has been updated in the last 2 days
            return S_OK()
        if not os.access(caDir, os.W_OK):
            gLogger.error("Your CRLs appear to be outdated, but you have no access to update them.")
            # Try to continue anyway...
            return S_OK()
        # Update the CAs & CRLs
        gLogger.notice("Your CRLs appear to be outdated; attempting to update them...")
        bdc = BundleDeliveryClient()
        res = bdc.syncCAs()
        if not res["OK"]:
            gLogger.error("Failed to update CAs", res["Message"])
        res = bdc.syncCRLs()
        if not res["OK"]:
            gLogger.error("Failed to update CRLs", res["Message"])
        # Continue even if the update failed...
        return S_OK()

    def doTheMagic(self):
        proxy = self.createProxy()
        if not proxy["OK"]:
            return proxy

        self.checkCAs()
        pI.certLifeTimeCheck()
        resultProxyWithVOMS = pI.addVOMSExtIfNeeded()
        if not resultProxyWithVOMS["OK"]:
            if "returning a valid AC for the user" in resultProxyWithVOMS["Message"]:
                gLogger.error(resultProxyWithVOMS["Message"])
                gLogger.error("\n Are you sure you are properly registered in the VO?")
            elif "Missing voms-proxy" in resultProxyWithVOMS["Message"]:
                gLogger.notice("Failed to add VOMS extension: no standard grid interface available")
            else:
                gLogger.error(resultProxyWithVOMS["Message"])
            if self.__piParams.strict:
                return resultProxyWithVOMS

        if self.__piParams.uploadProxy:
            resultProxyUpload = pI.uploadProxy()
            if not resultProxyUpload["OK"]:
                if self.__piParams.strict:
                    return resultProxyUpload

        return S_OK()


@Script()
def main():
    global piParams, pI
    piParams = Params()
    piParams.registerCLISwitches()

    Script.disableCS()
    Script.parseCommandLine(ignoreErrors=True)
    DIRAC.gConfig.setOptionValue("/DIRAC/Security/UseServerCertificate", "False")

    pI = ProxyInit(piParams)
    resultDoTheMagic = pI.doTheMagic()
    if not resultDoTheMagic["OK"]:
        gLogger.fatal(resultDoTheMagic["Message"])
        sys.exit(1)

    pI.printInfo()

    sys.exit(0)


if __name__ == "__main__":
    main()



""" Storage plug-ins related utilities
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import errno

from DIRAC import S_OK, S_ERROR


def checkArgumentFormat(path):
    """returns {'/this/is/an/lfn.1':False, '/this/is/an/lfn.2':False ...}"""

    if isinstance(path, six.string_types):
        return S_OK({path: False})
    elif isinstance(path, list):
        return S_OK(dict([(url, False) for url in path if isinstance(url, six.string_types)]))
    elif isinstance(path, dict):
        returnDict = path.copy()
        return S_OK(returnDict)
    else:
        return S_ERROR(errno.EINVAL, "Utils.checkArgumentFormat: Supplied path is not of the correct format.")



#!/usr/bin/env python

"""
Script that emulates the behaviour of a shell to edit the CS config.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import sys

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

# Invariants:
# * root does not end with "/" or root is "/"
# * root starts with "/"


@Script()
def main():
    Script.parseCommandLine()
    from DIRAC.ConfigurationSystem.Client.CSShellCLI import CSShellCLI

    shell = CSShellCLI()
    shell.cmdloop()


if __name__ == "__main__":
    main()



""" This contains unit tests to make sure that the migration between PyGSI and M2Crypto is as smooth as possible
    The test covers only the method exposed by the PyGSI version for the time being.

    We are not testing:

    * generateProxyRequest with bitStrengsh -> I should...
    * getIssuerDN: does not make any sense here, is never used
    * generateChainFromResponse: not used


    We are skipping:
    * init with arguments, because never used
    * The delegation mechanism (checkChain method)involves also X509Chain so it is in the X509Chain



"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


# redefined-outer-name is needed because we keep bassing get_X509Chain_class as param
# pylint: disable=redefined-outer-name

from datetime import datetime, timedelta
from string import ascii_letters


from pytest import mark, fixture, skip, raises, approx

parametrize = mark.parametrize

from DIRAC.Core.Security.test.x509TestUtilities import (
    CERTS,
    CERTKEYS,
    CERTCONTENTS,
    deimportDIRAC,
    ENCRYPTEDKEYPASS,
    ENCRYPTEDKEY,
    getCertOption,
    HOSTCERT,
    KEYCONTENTS_PKCS8,
    USERCERT,
    X509REQUESTTYPES,
    get_X509Request,
)


def test_dumpRequest_notInitialized(get_X509Request):
    """Calls dumpRequest a non initlaized Request"""

    x509Req = get_X509Request()
    res = x509Req.dumpRequest()

    assert res["OK"] is False
    from DIRAC.Core.Utilities.DErrno import ENOCERT

    assert res["Errno"] == ENOCERT


def test_dumpRequest(get_X509Request):
    """ " Generate an X509Request and dumps it"""
    x509Req = get_X509Request()
    x509Req.generateProxyRequest()

    res = x509Req.dumpRequest()

    assert res["OK"]
    assert b"CERTIFICATE REQUEST" in res["Value"]


def test_loadAllFromString_fromDumpRequest(get_X509Request):
    """Generate a proxy Request and try loading it from incomplete dump"""
    x509Req = get_X509Request()
    x509Req.generateProxyRequest()

    proxyRequest = x509Req.dumpRequest()["Value"]

    # This should fail because the proxyRequest does not contain the private key
    x509ReqLoad = get_X509Request()
    res = x509ReqLoad.loadAllFromString(proxyRequest)

    assert res["OK"] is False
    from DIRAC.Core.Utilities.DErrno import ENOPKEY

    assert res["Errno"] == ENOPKEY


@parametrize("isLimited", (False, True))
def test_getSubjectDN(get_X509Request, isLimited):
    """Try getting the subjectDN in case of limited and non limited request
    :param isLimited: request a limited proxy
    """

    x509Req = get_X509Request()
    x509Req.generateProxyRequest(limited=isLimited)

    res = x509Req.getSubjectDN()
    assert res["OK"]

    if isLimited:
        assert res["Value"] == "/CN=limited proxy"
    else:
        assert res["Value"] == "/CN=proxy"


@parametrize("isLimited", (False, True))
def test_loadAllFromString(get_X509Request, isLimited):
    """Generate a proxy Request, load it, and check that the subject DN are the same
    :param isLimited: request a limited proxy"""
    x509Req = get_X509Request()
    x509Req.generateProxyRequest(limited=isLimited)

    proxyRequest = x509Req.dumpAll()["Value"]

    x509ReqLoad = get_X509Request()
    res = x509ReqLoad.loadAllFromString(proxyRequest)

    assert res["OK"]
    assert x509Req.getSubjectDN() == x509ReqLoad.getSubjectDN()



""" Unit Test of Workflow Modules
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
import itertools
import os
import copy
import shutil

from mock import MagicMock as Mock

from DIRAC import gLogger


class ModulesTestCase(unittest.TestCase):
    """Base class for the Modules test cases"""

    def setUp(self):

        gLogger.setLevel("ERROR")
        #    import sys
        #    sys.modules["DIRAC"] = DIRAC.ResourceStatusSystem.test.fake_Logger
        #    sys.modules["DIRAC.ResourceStatusSystem.Utilities.CS"] = DIRAC.ResourceStatusSystem.test.fake_Logger

        self.jr_mock = Mock()
        self.jr_mock.setApplicationStatus.return_value = {"OK": True, "Value": ""}
        self.jr_mock.generateRequest.return_value = {"OK": True, "Value": "pippo"}
        self.jr_mock.setJobParameter.return_value = {"OK": True, "Value": "pippo"}
        self.jr_mock.generateForwardDISET.return_value = {"OK": True, "Value": "pippo"}
        #    self.jr_mock.setJobApplicationStatus.return_value = {'OK': True, 'Value': 'pippo'}

        self.fr_mock = Mock()
        self.fr_mock.getFiles.return_value = {}
        self.fr_mock.setFileStatus.return_value = {"OK": True, "Value": ""}
        self.fr_mock.commit.return_value = {"OK": True, "Value": ""}
        self.fr_mock.generateRequest.return_value = {"OK": True, "Value": ""}

        rc_mock = Mock()
        rc_mock.update.return_value = {"OK": True, "Value": ""}
        rc_mock.setDISETRequest.return_value = {"OK": True, "Value": ""}
        rc_mock.isEmpty.return_value = {"OK": True, "Value": ""}
        rc_mock.toXML.return_value = {"OK": True, "Value": ""}
        rc_mock.getDigest.return_value = {"OK": True, "Value": ""}
        rc_mock.__len__.return_value = 1
        self.rc_mock = rc_mock

        ar_mock = Mock()
        ar_mock.commit.return_value = {"OK": True, "Value": ""}

        self.rm_mock = Mock()
        self.rm_mock.getReplicas.return_value = {
            "OK": True,
            "Value": {"Successful": {"pippo": "metadataPippo"}, "Failed": None},
        }
        self.rm_mock.getCatalogFileMetadata.return_value = {
            "OK": True,
            "Value": {"Successful": {"pippo": "metadataPippo"}, "Failed": None},
        }
        self.rm_mock.removeFile.return_value = {"OK": True, "Value": {"Failed": False}}
        self.rm_mock.putStorageDirectory.return_value = {"OK": True, "Value": {"Failed": False}}
        self.rm_mock.addCatalogFile.return_value = {"OK": True, "Value": {"Failed": False}}
        self.rm_mock.putAndRegister.return_value = {"OK": True, "Value": {"Failed": False}}
        self.rm_mock.getFile.return_value = {"OK": True, "Value": {"Failed": False}}

        self.jsu_mock = Mock()
        self.jsu_mock.setJobApplicationStatus.return_value = {"OK": True, "Value": ""}

        self.jsu_mock = Mock()
        self.jsu_mock.setJobApplicationStatus.return_value = {"OK": True, "Value": ""}

        request_mock = Mock()
        request_mock.addSubRequest.return_value = {"OK": True, "Value": ""}
        request_mock.setSubRequestFiles.return_value = {"OK": True, "Value": ""}
        request_mock.getNumSubRequests.return_value = {"OK": True, "Value": ""}
        request_mock._getLastOrder.return_value = 1

        self.ft_mock = Mock()
        self.ft_mock.transferAndRegisterFile.return_value = {"OK": True, "Value": {"uploadedSE": ""}}
        self.ft_mock.transferAndRegisterFileFailover.return_value = {"OK": True, "Value": {}}

        self.nc_mock = Mock()
        self.nc_mock.sendMail.return_value = {"OK": True, "Value": ""}

        self.prod_id = 123
        self.prod_job_id = 456
        self.wms_job_id = 0
        self.workflowStatus = {"OK": True}
        self.stepStatus = {"OK": True}

        self.wf_commons = [
            {
                "PRODUCTION_ID": str(self.prod_id),
                "JOB_ID": str(self.prod_job_id),
                "eventType": "123456789",
                "jobType": "merge",
                "configName": "aConfigName",
                "configVersion": "aConfigVersion",
                "outputDataFileMask": "",
                "BookkeepingLFNs": "aa",
                "ProductionOutputData": "ProductionOutputData",
                "numberOfEvents": "100",
                "JobReport": self.jr_mock,
                "Request": rc_mock,
                "AccountingReport": ar_mock,
                "FileReport": self.fr_mock,
                "runNumber": "Unknown",
                "appSteps": ["someApp_1"],
            },
            {
                "PRODUCTION_ID": str(self.prod_id),
                "JOB_ID": str(self.prod_job_id),
                "configName": "aConfigName",
                "configVersion": "aConfigVersion",
                "outputDataFileMask": "",
                "jobType": "merge",
                "BookkeepingLFNs": "aa",
                "ProductionOutputData": "ProductionOutputData",
                "numberOfEvents": "100",
                "JobReport": self.jr_mock,
                "Request": rc_mock,
                "AccountingReport": ar_mock,
                "FileReport": self.fr_mock,
                "LogFilePath": "someDir",
                "runNumber": "Unknown",
                "appSteps": ["someApp_1"],
            },
            {
                "PRODUCTION_ID": str(self.prod_id),
                "JOB_ID": str(self.prod_job_id),
                "configName": "aConfigName",
                "configVersion": "aConfigVersion",
                "outputDataFileMask": "",
                "jobType": "merge",
                "BookkeepingLFNs": "aa",
                "ProductionOutputData": "ProductionOutputData",
                "numberOfEvents": "100",
                "JobReport": self.jr_mock,
                "Request": rc_mock,
                "AccountingReport": ar_mock,
                "FileReport": self.fr_mock,
                "LogFilePath": "someDir",
                "LogTargetPath": "someOtherDir",
                "runNumber": "Unknown",
                "appSteps": ["someApp_1"],
            },
            {
                "PRODUCTION_ID": str(self.prod_id),
                "JOB_ID": str(self.prod_job_id),
                "configName": "aConfigName",
                "configVersion": "aConfigVersion",
                "outputDataFileMask": "",
                "jobType": "merge",
                "BookkeepingLFNs": "aa",
                "ProductionOutputData": "ProductionOutputData",
                "numberOfEvents": "100",
                "JobReport": self.jr_mock,
                "Request": rc_mock,
                "AccountingReport": ar_mock,
                "FileReport": self.fr_mock,
                "LogFilePath": "someDir",
                "LogTargetPath": "someOtherDir",
                "runNumber": "Unknown",
                "appSteps": ["someApp_1"],
            },
            {
                "PRODUCTION_ID": str(self.prod_id),
                "JOB_ID": str(self.prod_job_id),
                "configName": "aConfigName",
                "configVersion": "aConfigVersion",
                "outputDataFileMask": "",
                "jobType": "reco",
                "BookkeepingLFNs": "aa",
                "ProductionOutputData": "ProductionOutputData",
                "JobReport": self.jr_mock,
                "Request": rc_mock,
                "AccountingReport": ar_mock,
                "FileReport": self.fr_mock,
                "runNumber": "Unknown",
                "appSteps": ["someApp_1"],
            },
            {
                "PRODUCTION_ID": str(self.prod_id),
                "JOB_ID": str(self.prod_job_id),
                "configName": "aConfigName",
                "configVersion": "aConfigVersion",
                "outputDataFileMask": "",
                "jobType": "reco",
                "BookkeepingLFNs": "aa",
                "ProductionOutputData": "ProductionOutputData",
                "JobReport": self.jr_mock,
                "Request": rc_mock,
                "AccountingReport": ar_mock,
                "FileReport": self.fr_mock,
                "LogFilePath": "someDir",
                "runNumber": "Unknown",
                "appSteps": ["someApp_1"],
            },
            {
                "PRODUCTION_ID": str(self.prod_id),
                "JOB_ID": str(self.prod_job_id),
                "configName": "aConfigName",
                "configVersion": "aConfigVersion",
                "outputDataFileMask": "",
                "jobType": "reco",
                "BookkeepingLFNs": "aa",
                "ProductionOutputData": "ProductionOutputData",
                "JobReport": self.jr_mock,
                "Request": rc_mock,
                "AccountingReport": ar_mock,
                "FileReport": self.fr_mock,
                "LogFilePath": "someDir",
                "LogTargetPath": "someOtherDir",
                "runNumber": "Unknown",
                "appSteps": ["someApp_1"],
            },
            {
                "PRODUCTION_ID": str(self.prod_id),
                "JOB_ID": str(self.prod_job_id),
                "configName": "aConfigName",
                "configVersion": "aConfigVersion",
                "outputDataFileMask": "",
                "jobType": "reco",
                "BookkeepingLFNs": "aa",
                "ProductionOutputData": "ProductionOutputData",
                "JobReport": self.jr_mock,
                "Request": rc_mock,
                "AccountingReport": ar_mock,
                "FileReport": self.fr_mock,
                "LogFilePath": "someDir",
                "LogTargetPath": "someOtherDir",
                "runNumber": "Unknown",
                "appSteps": ["someApp_1"],
            },
            {
                "PRODUCTION_ID": str(self.prod_id),
                "JOB_ID": str(self.prod_job_id),
                "configName": "aConfigName",
                "configVersion": "aConfigVersion",
                "outputDataFileMask": "",
                "jobType": "reco",
                "BookkeepingLFNs": "aa",
                "ProductionOutputData": "ProductionOutputData",
                "JobReport": self.jr_mock,
                "Request": rc_mock,
                "AccountingReport": ar_mock,
                "FileReport": self.fr_mock,
                "LogFilePath": "someDir",
                "LogTargetPath": "someOtherDir",
                "runNumber": "Unknown",
                "InputData": "",
                "appSteps": ["someApp_1"],
            },
            {
                "PRODUCTION_ID": str(self.prod_id),
                "JOB_ID": str(self.prod_job_id),
                "configName": "aConfigName",
                "configVersion": "aConfigVersion",
                "outputDataFileMask": "",
                "jobType": "reco",
                "BookkeepingLFNs": "aa",
                "ProductionOutputData": "ProductionOutputData",
                "JobReport": self.jr_mock,
                "Request": rc_mock,
                "AccountingReport": ar_mock,
                "FileReport": self.fr_mock,
                "LogFilePath": "someDir",
                "LogTargetPath": "someOtherDir",
                "runNumber": "Unknown",
                "InputData": "foo;bar",
                "appSteps": ["someApp_1"],
            },
            {
                "PRODUCTION_ID": str(self.prod_id),
                "JOB_ID": str(self.prod_job_id),
                "configName": "aConfigName",
                "configVersion": "aConfigVersion",
                "outputDataFileMask": "",
                "jobType": "reco",
                "BookkeepingLFNs": "aa",
                "ProductionOutputData": "ProductionOutputData",
                "JobReport": self.jr_mock,
                "Request": rc_mock,
                "AccountingReport": ar_mock,
                "FileReport": self.fr_mock,
                "LogFilePath": "someDir",
                "LogTargetPath": "someOtherDir",
                "runNumber": "Unknown",
                "InputData": "foo;bar",
                "ParametricInputData": "",
                "appSteps": ["someApp_1"],
            },
            {
                "PRODUCTION_ID": str(self.prod_id),
                "JOB_ID": str(self.prod_job_id),
                "configName": "aConfigName",
                "configVersion": "aConfigVersion",
                "outputDataFileMask": "",
                "jobType": "reco",
                "BookkeepingLFNs": "aa",
                "ProductionOutputData": "ProductionOutputData",
                "JobReport": self.jr_mock,
                "Request": rc_mock,
                "AccountingReport": ar_mock,
                "FileReport": self.fr_mock,
                "LogFilePath": "someDir",
                "LogTargetPath": "someOtherDir",
                "runNumber": "Unknown",
                "InputData": "foo;bar",
                "ParametricInputData": "pid1;pid2;pid3",
                "appSteps": ["someApp_1"],
            },
        ]
        self.step_commons = [
            {
                "applicationName": "someApp",
                "applicationVersion": "v1r0",
                "eventType": "123456789",
                "applicationLog": "appLog",
                "extraPackages": "",
                "XMLSummary": "XMLSummaryFile",
                "numberOfEvents": "100",
                "BKStepID": "123",
                "StepProcPass": "Sim123",
                "outputFilePrefix": "pref_",
                "STEP_INSTANCE_NAME": "someApp_1",
                "listoutput": [
                    {
                        "outputDataName": str(self.prod_id) + "_" + str(self.prod_job_id) + "_",
                        "outputDataSE": "aaa",
                        "outputDataType": "bbb",
                    }
                ],
            },
            {
                "applicationName": "someApp",
                "applicationVersion": "v1r0",
                "eventType": "123456789",
                "applicationLog": "appLog",
                "extraPackages": "",
                "XMLSummary": "XMLSummaryFile",
                "numberOfEvents": "100",
                "BKStepID": "123",
                "StepProcPass": "Sim123",
                "outputFilePrefix": "pref_",
                "optionsLine": "",
                "STEP_INSTANCE_NAME": "someApp_1",
                "listoutput": [
                    {
                        "outputDataName": str(self.prod_id) + "_" + str(self.prod_job_id) + "_",
                        "outputDataSE": "aaa",
                        "outputDataType": "bbb",
                    }
                ],
            },
            {
                "applicationName": "someApp",
                "applicationVersion": "v1r0",
                "eventType": "123456789",
                "applicationLog": "appLog",
                "extraPackages": "",
                "XMLSummary": "XMLSummaryFile",
                "numberOfEvents": "100",
                "BKStepID": "123",
                "StepProcPass": "Sim123",
                "outputFilePrefix": "pref_",
                "extraOptionsLine": "blaBla",
                "STEP_INSTANCE_NAME": "someApp_1",
                "listoutput": [
                    {
                        "outputDataName": str(self.prod_id) + "_" + str(self.prod_job_id) + "_",
                        "outputDataSE": "aaa",
                        "outputDataType": "bbb",
                    }
                ],
            },
        ]
        self.step_number = "321"
        self.step_id = "%s_%s_%s" % (self.prod_id, self.prod_job_id, self.step_number)

        from DIRAC.Workflow.Modules.ModuleBase import ModuleBase

        self.mb = ModuleBase()

        self.mb.rm = self.rm_mock
        self.mb.request = self.rc_mock
        self.mb.jobReport = self.jr_mock
        self.mb.fileReport = self.fr_mock
        self.mb.workflow_commons = self.wf_commons[0]

        from DIRAC.Workflow.Modules.FailoverRequest import FailoverRequest

        self.fr = FailoverRequest()

        self.fr.request = self.rc_mock
        self.fr.jobReport = self.jr_mock
        self.fr.fileReport = self.fr_mock

        from DIRAC.Workflow.Modules.Script import Script

        self.script = Script()

        self.script.request = self.rc_mock
        self.script.jobReport = self.jr_mock
        self.script.fileReport = self.fr_mock

    def tearDown(self):
        for fileProd in [
            "appLog",
            "foo.txt",
            "aaa.Bhadron.dst",
            "bbb.Calibration.dst",
            "bar_2.py",
            "foo_1.txt",
            "ccc.charm.mdst",
            "prova.txt",
            "foo.txt",
            "BAR.txt",
            "FooBAR.ext.txt",
            "applicationLog.txt",
            "ErrorLogging_Step1_coredump.log",
            "123_00000456_request.xml",
            "lfn1",
            "lfn2",
            "aaa.bhadron.dst",
            "bbb.calibration.dst",
            "ProductionOutputData",
            "data.py",
            "00000123_00000456.tar",
            "someOtherDir",
            "DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK",
        ]:
            try:
                os.remove(fileProd)
            except OSError:
                continue

        for directory in ["./job", "job"]:
            try:
                shutil.rmtree(directory)
            except Exception:
                continue


#############################################################################
# ModuleBase.py
#############################################################################


class ModuleBaseSuccess(ModulesTestCase):

    #################################################

    def test__checkLocalExistance(self):

        self.assertRaises(OSError, self.mb._checkLocalExistance, ["aaa", "bbb"])

    #################################################

    def test__applyMask(self):

        candidateFiles = {
            "00012345_00012345_4.dst": {
                "lfn": "/lhcb/MC/2010/DST/00012345/0001/00012345_00012345_4.dst",
                "type": "dst",
                "workflowSE": "Tier1_MC_M-DST",
            },
            "00012345_00012345_2.digi": {"type": "digi", "workflowSE": "Tier1-RDST"},
            "00012345_00012345_3.digi": {"type": "digi", "workflowSE": "Tier1-RDST"},
            "00012345_00012345_5.AllStreams.dst": {
                "lfn": "/lhcb/MC/2010/DST/00012345/0001/00012345_00012345_5.AllStreams.dst",
                "type": "allstreams.dst",
                "workflowSE": "Tier1_MC_M-DST",
            },
            "00012345_00012345_1.sim": {"type": "sim", "workflowSE": "Tier1-RDST"},
        }

        fileMasks = (["dst"], "dst", ["sim"], ["digi"], ["digi", "sim"], "allstreams.dst")
        stepMasks = ("", "5", "", ["2"], ["1", "3"], "")

        results = (
            {
                "00012345_00012345_4.dst": {
                    "lfn": "/lhcb/MC/2010/DST/00012345/0001/00012345_00012345_4.dst",
                    "type": "dst",
                    "workflowSE": "Tier1_MC_M-DST",
                }
            },
            {},
            {"00012345_00012345_1.sim": {"type": "sim", "workflowSE": "Tier1-RDST"}},
            {
                "00012345_00012345_2.digi": {"type": "digi", "workflowSE": "Tier1-RDST"},
            },
            {
                "00012345_00012345_3.digi": {"type": "digi", "workflowSE": "Tier1-RDST"},
                "00012345_00012345_1.sim": {"type": "sim", "workflowSE": "Tier1-RDST"},
            },
            {
                "00012345_00012345_5.AllStreams.dst": {
                    "lfn": "/lhcb/MC/2010/DST/00012345/0001/00012345_00012345_5.AllStreams.dst",
                    "type": "allstreams.dst",
                    "workflowSE": "Tier1_MC_M-DST",
                }
            },
        )

        for fileMask, result, stepMask in zip(fileMasks, results, stepMasks):
            res = self.mb._applyMask(candidateFiles, fileMask, stepMask)
            self.assertEqual(res, result)

    #################################################

    def test__checkSanity(self):

        candidateFiles = {
            "00012345_00012345_4.dst": {
                "lfn": "/lhcb/MC/2010/DST/00012345/0001/00012345_00012345_4.dst",
                "type": "dst",
                "workflowSE": "Tier1_MC_M-DST",
            },
            "00012345_00012345_2.digi": {"type": "digi", "workflowSE": "Tier1-RDST"},
            "00012345_00012345_3.digi": {"type": "digi", "workflowSE": "Tier1-RDST"},
            "00012345_00012345_5.AllStreams.dst": {
                "lfn": "/lhcb/MC/2010/DST/00012345/0001/00012345_00012345_5.AllStreams.dst",
                "type": "DST",
                "workflowSE": "Tier1_MC_M-DST",
            },
            "00012345_00012345_1.sim": {"type": "sim", "workflowSE": "Tier1-RDST"},
        }

        self.assertRaises(ValueError, self.mb._checkSanity, candidateFiles)

    #################################################

    def test_getCandidateFiles(self):
        # this needs to avoid the "checkLocalExistance"

        open("foo_1.txt", "w").close()
        open("bar_2.py", "w").close()

        outputList = [
            {"outputDataType": "txt", "outputDataSE": "Tier1-RDST", "outputDataName": "foo_1.txt"},
            {"outputDataType": "py", "outputDataSE": "Tier1-RDST", "outputDataName": "bar_2.py"},
        ]
        outputLFNs = ["/lhcb/MC/2010/DST/00012345/0001/foo_1.txt", "/lhcb/MC/2010/DST/00012345/0001/bar_2.py"]
        fileMask = "txt"
        stepMask = ""
        result = {
            "foo_1.txt": {
                "lfn": "/lhcb/MC/2010/DST/00012345/0001/foo_1.txt",
                "type": outputList[0]["outputDataType"],
                "workflowSE": outputList[0]["outputDataSE"],
            }
        }

        res = self.mb.getCandidateFiles(outputList, outputLFNs, fileMask, stepMask)
        self.assertEqual(res, result)

        fileMask = ["txt", "py"]
        stepMask = None
        result = {
            "foo_1.txt": {
                "lfn": "/lhcb/MC/2010/DST/00012345/0001/foo_1.txt",
                "type": outputList[0]["outputDataType"],
                "workflowSE": outputList[0]["outputDataSE"],
            },
            "bar_2.py": {
                "lfn": "/lhcb/MC/2010/DST/00012345/0001/bar_2.py",
                "type": outputList[1]["outputDataType"],
                "workflowSE": outputList[1]["outputDataSE"],
            },
        }
        res = self.mb.getCandidateFiles(outputList, outputLFNs, fileMask, stepMask)
        self.assertEqual(res, result)

        fileMask = ["aa"]
        stepMask = None
        res = self.mb.getCandidateFiles(outputList, outputLFNs, fileMask, stepMask)
        result = {}
        self.assertEqual(res, result)

        fileMask = ""
        stepMask = "2"
        result = {
            "bar_2.py": {
                "lfn": "/lhcb/MC/2010/DST/00012345/0001/bar_2.py",
                "type": outputList[1]["outputDataType"],
                "workflowSE": outputList[1]["outputDataSE"],
            }
        }

        res = self.mb.getCandidateFiles(outputList, outputLFNs, fileMask, stepMask)

        self.assertEqual(res, result)

        fileMask = ""
        stepMask = 2
        result = {
            "bar_2.py": {
                "lfn": "/lhcb/MC/2010/DST/00012345/0001/bar_2.py",
                "type": outputList[1]["outputDataType"],
                "workflowSE": outputList[1]["outputDataSE"],
            }
        }

        res = self.mb.getCandidateFiles(outputList, outputLFNs, fileMask, stepMask)

        self.assertEqual(res, result)

        fileMask = ""
        stepMask = ["2", "3"]
        result = {
            "bar_2.py": {
                "lfn": "/lhcb/MC/2010/DST/00012345/0001/bar_2.py",
                "type": outputList[1]["outputDataType"],
                "workflowSE": outputList[1]["outputDataSE"],
            }
        }

        res = self.mb.getCandidateFiles(outputList, outputLFNs, fileMask, stepMask)

        self.assertEqual(res, result)

        fileMask = ""
        stepMask = ["3"]
        result = {}

        res = self.mb.getCandidateFiles(outputList, outputLFNs, fileMask, stepMask)

        self.assertEqual(res, result)

    def test__enableModule(self):

        self.mb.production_id = self.prod_id
        self.mb.prod_job_id = self.prod_job_id
        self.mb.jobID = self.wms_job_id
        self.mb.workflowStatus = self.workflowStatus
        self.mb.stepStatus = self.stepStatus
        self.mb.workflow_commons = self.wf_commons[0]  # APS: this is needed
        self.mb.step_commons = self.step_commons[0]
        self.mb.step_number = self.step_number
        self.mb.step_id = self.step_id
        self.mb.execute()
        self.assertFalse(self.mb._enableModule())

        self.mb.jobID = 1
        self.mb.execute()
        self.assertTrue(self.mb._enableModule())

    def test__determineStepInputData(self):

        self.mb.stepName = "DaVinci_2"

        inputData = "previousStep"
        self.mb.appSteps = ["Brunel_1", "DaVinci_2"]
        self.mb.workflow_commons = {
            "outputList": [
                {
                    "stepName": "Brunel_1",
                    "outputDataType": "brunelhist",
                    "outputBKType": "BRUNELHIST",
                    "outputDataSE": "CERN-HIST",
                    "outputDataName": "Brunel_00012345_00006789_1_Hist.root",
                },
                {
                    "stepName": "Brunel_1",
                    "outputDataType": "sdst",
                    "outputBKType": "SDST",
                    "outputDataSE": "Tier1-BUFFER",
                    "outputDataName": "00012345_00006789_1.sdst",
                },
            ]
        }
        self.mb.inputDataType = "SDST"

        first = self.mb._determineStepInputData(inputData)
        second = ["00012345_00006789_1.sdst"]
        self.assertEqual(first, second)

        inputData = "previousStep"
        self.mb.appSteps = ["Brunel_1", "DaVinci_2"]
        self.mb.workflow_commons["outputList"] = [
            {
                "stepName": "Brunel_1",
                "outputDataType": "brunelhist",
                "outputBKType": "BRUNELHIST",
                "outputDataSE": "CERN-HIST",
                "outputDataName": "Brunel_00012345_00006789_1_Hist.root",
            },
            {
                "stepName": "Brunel_1",
                "outputDataType": "sdst",
                "outputBKType": "SDST",
                "outputDataSE": "Tier1-BUFFER",
                "outputDataName": "some.sdst",
            },
            {
                "stepName": "Brunel_1",
                "outputDataType": "sdst",
                "outputBKType": "SDST",
                "outputDataSE": "Tier1-BUFFER",
                "outputDataName": "00012345_00006789_1.sdst",
            },
        ]
        self.mb.inputDataType = "SDST"
        first = self.mb._determineStepInputData(inputData)
        second = ["some.sdst", "00012345_00006789_1.sdst"]
        self.assertEqual(first, second)

        inputData = "LFN:123.raw"
        first = self.mb._determineStepInputData(inputData)
        second = ["123.raw"]
        self.assertEqual(first, second)


#############################################################################
# FailoverRequest.py
#############################################################################


class FailoverRequestSuccess(ModulesTestCase):

    #################################################

    def test_execute(self):

        self.fr.jobType = "merge"
        self.fr.stepInputData = ["foo", "bar"]

        self.fr.production_id = self.prod_id
        self.fr.prod_job_id = self.prod_job_id
        self.fr.jobID = self.wms_job_id
        self.fr.workflowStatus = self.workflowStatus
        self.fr.stepStatus = self.stepStatus
        self.fr.workflow_commons = self.wf_commons
        self.fr.step_commons = self.step_commons[0]
        self.fr.step_number = self.step_number
        self.fr.step_id = self.step_id

        # no errors, no input data
        for wf_commons in copy.deepcopy(self.wf_commons):
            for step_commons in self.step_commons:
                self.fr.workflow_commons = wf_commons
                self.fr.step_commons = step_commons
                res = self.fr.execute()
                self.assertTrue(res["OK"])


#############################################################################
# Scripy.py
#############################################################################


class ScriptSuccess(ModulesTestCase):

    #################################################

    def test_execute(self):

        self.script.jobType = "merge"
        self.script.stepInputData = ["foo", "bar"]

        self.script.production_id = self.prod_id
        self.script.prod_job_id = self.prod_job_id
        self.script.jobID = self.wms_job_id
        self.script.workflowStatus = self.workflowStatus
        self.script.stepStatus = self.stepStatus
        self.script.workflow_commons = self.wf_commons
        self.script.step_commons = self.step_commons[0]
        self.script.step_number = self.step_number
        self.script.step_id = self.step_id
        self.script.executable = "ls"
        self.script.applicationLog = "applicationLog.txt"

        # no errors, no input data
        for wf_commons in copy.deepcopy(self.wf_commons):
            for step_commons in self.step_commons:
                self.script.workflow_commons = wf_commons
                self.script.step_commons = step_commons
                self.script._setCommand()
                self.script._executeCommand()


class ScriptUnicode(ModulesTestCase):

    #################################################

    def test_execute(self):

        self.script.jobType = "merge"
        self.script.stepInputData = ["foo", "bar"]

        self.script.production_id = self.prod_id
        self.script.prod_job_id = self.prod_job_id
        self.script.jobID = self.wms_job_id
        self.script.workflowStatus = self.workflowStatus
        self.script.stepStatus = self.stepStatus
        self.script.workflow_commons = self.wf_commons
        self.script.step_commons = self.step_commons[0]
        self.script.step_number = self.step_number
        self.script.step_id = self.step_id
        self.script.executable = "echo"
        self.script.arguments = r"-e '\xe2\x96\x88\xe2\x96\x88\xe2\x96\x88\xe2\x96\x88\xe2\x96\x88\xe2\x95\x97'"
        self.script.applicationLog = "applicationLog.txt"

        # no errors, no input data
        for wf_commons in copy.deepcopy(self.wf_commons):
            for step_commons in self.step_commons:
                self.script.workflow_commons = wf_commons
                self.script.step_commons = step_commons
                self.script._setCommand()
                self.script._executeCommand()


class ScriptFailure(ModulesTestCase):

    #################################################

    def test_execute(self):

        self.script.jobType = "merge"
        self.script.stepInputData = ["foo", "bar"]

        self.script.production_id = self.prod_id
        self.script.prod_job_id = self.prod_job_id
        self.script.jobID = self.wms_job_id
        self.script.workflowStatus = self.workflowStatus
        self.script.stepStatus = self.stepStatus
        self.script.workflow_commons = self.wf_commons
        self.script.step_commons = self.step_commons[0]
        self.script.step_number = self.step_number
        self.script.step_id = self.step_id

        # no errors, no input data
        for wf_commons in copy.deepcopy(self.wf_commons):
            for step_commons in self.step_commons:
                self.script.workflow_commons = wf_commons
                self.script.step_commons = step_commons
                res = self.script.execute()
                self.assertFalse(res["OK"])


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(ModulesTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ModuleBaseSuccess))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(FailoverRequestSuccess))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ScriptSuccess))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ScriptFailure))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



#!/usr/bin/env python
"""
Add or Modify a User info in DIRAC

Example:
  $ dirac-admin-add-user -N vhamar -D /O=GRID/C=FR/O=CNRS/OU=CPPM/CN=Vanessa Hamar -M hamar@cppm.in2p3.fr -G dirac_user
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

userName = None
userDN = None
userMail = None
userGroups = []


def setUserName(arg):
    global userName
    if userName or not arg:
        Script.showHelp(exitCode=1)
    userName = arg


def setUserDN(arg):
    global userDN
    if userDN or not arg:
        Script.showHelp(exitCode=1)
    userDN = arg


def setUserMail(arg):
    global userMail
    if userMail or not arg:
        Script.showHelp(exitCode=1)
    if not arg.find("@") > 0:
        gLogger.error("Not a valid mail address", arg)
        DIRAC.exit(-1)
    userMail = arg


def addUserGroup(arg):
    global userGroups
    if not arg:
        Script.showHelp(exitCode=1)
    if arg not in userGroups:
        userGroups.append(arg)


@Script()
def main():
    global userName
    global userDN
    global userMail
    global userGroups
    Script.registerSwitch("N:", "UserName:", "Short Name of the User (Mandatory)", setUserName)
    Script.registerSwitch("D:", "UserDN:", "DN of the User Certificate (Mandatory)", setUserDN)
    Script.registerSwitch("M:", "UserMail:", "eMail of the user (Mandatory)", setUserMail)
    Script.registerSwitch(
        "G:", "UserGroup:", "Name of the Group for the User (Allow Multiple instances or None)", addUserGroup
    )
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["Property=<Value>: Properties to be added to the User like (Phone=XXXX)"], mandatory=False)
    Script.parseCommandLine(ignoreErrors=True)

    if userName is None or userDN is None or userMail is None:
        Script.showHelp(exitCode=1)

    args = Script.getPositionalArgs()

    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    errorList = []

    userProps = {"DN": userDN, "Email": userMail}
    if userGroups:
        userProps["Groups"] = userGroups
    for prop in args:
        pl = prop.split("=")
        if len(pl) < 2:
            errorList.append(("in arguments", "Property %s has to include a '=' to separate name from value" % prop))
            exitCode = 255
        else:
            pName = pl[0]
            pValue = "=".join(pl[1:])
            gLogger.info("Setting property %s to %s" % (pName, pValue))
            userProps[pName] = pValue

    if not diracAdmin.csModifyUser(userName, userProps, createIfNonExistant=True)["OK"]:
        errorList.append(("add user", "Cannot register user %s" % userName))
        exitCode = 255
    else:
        result = diracAdmin.csCommitChanges()
        if not result["OK"]:
            errorList.append(("commit", result["Message"]))
            exitCode = 255

    for error in errorList:
        gLogger.error("%s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" Module that contains a mock object of the FileCatalog(s)
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=invalid-name, line-too-long

import datetime
from mock import MagicMock

from DIRAC import S_OK

fc_mock = MagicMock()

fc_mock.listDirectory.return_value = {
    "OK": True,
    "Value": {
        "Failed": {},
        "Successful": {
            "/this/is/dir1/": {
                "Datasets": {},
                "Files": {
                    "/this/is/dir1/file1.txt": {
                        "MetaData": {
                            "Checksum": "7149ed85",
                            "ChecksumType": "Adler32",
                            "CreationDate": datetime.datetime(2014, 12, 4, 12, 16, 56),
                            "FileID": 156301805,  # noqa
                            "GID": 2695,  # noqa
                            "GUID": "6A5C6C86-AD7B-E411-9EDB",
                            "Mode": 436,
                            "ModificationDate": datetime.datetime(2014, 12, 4, 12, 16, 56),
                            "Owner": "phicharp",
                            "OwnerGroup": "lhcb_prod",
                            "Size": 206380531,  # noqa
                            "Status": "AprioriGood",
                            "Type": "File",
                            "UID": 19503,  # noqa
                        }
                    },
                    "/this/is/dir1/file2.foo.bar": {
                        "MetaData": {
                            "Checksum": "7149ed86",
                            "ChecksumType": "Adler32",
                            "CreationDate": datetime.datetime(2014, 12, 4, 12, 16, 56),
                            "FileID": 156301805,  # noqa
                            "GID": 2695,  # noqa
                            "GUID": "6A5C6C86-AD7B-E411-9EDB",
                            "Mode": 436,
                            "ModificationDate": datetime.datetime(2014, 12, 4, 12, 16, 56),
                            "Owner": "phicharp",
                            "OwnerGroup": "lhcb_prod",
                            "Size": 206380532,  # noqa
                            "Status": "AprioriGood",
                            "Type": "File",
                            "UID": 19503,  # noqa
                        }
                    },
                },
                "Links": {},
                "SubDirs": {},
            },
            "/this/is/dir2/": {
                "Datasets": {},
                "Files": {},
                "Links": {},
                "SubDirs": {"/this/is/dir2/subdir1/": True, "/this/is/dir2/subdir2/": True},
            },
            "/this/is/dir2/subdir1/": {
                "Datasets": {},
                "Files": {
                    "/this/is/dir2/subdir1/file3.pippo": {
                        "MetaData": {
                            "Checksum": "7149ed86",
                            "ChecksumType": "Adler32",
                            "CreationDate": datetime.datetime(2014, 12, 4, 12, 16, 56),
                            "FileID": 156301805,  # noqa
                            "GID": 2695,  # noqa
                            "GUID": "6A5C6C86-AD7B-E411-9EDB",
                            "Mode": 436,
                            "ModificationDate": datetime.datetime(2014, 12, 4, 12, 16, 56),
                            "Owner": "phicharp",
                            "OwnerGroup": "lhcb_prod",
                            "Size": 206380532,  # noqa
                            "Status": "AprioriGood",
                            "Type": "File",
                            "UID": 19503,  # noqa
                        }
                    }
                },
                "Links": {},
                "SubDirs": {},
            },
            "/this/is/dir2/subdir2/": {"Datasets": {}, "Files": {}, "Links": {}, "SubDirs": {}},
        },
    },
}


fc_mock.getReplicas.return_value = {
    "OK": True,
    "Value": {
        "Failed": {},
        "Successful": {
            "/this/is/dir1/file1.txt": {
                "SE1": "smr://srm.SE1.ch:8443/srm/v2/server?SFN=/this/is/dir1/file1.txt",
                "SE2": "smr://srm.SE2.fr:8443/srm/v2/server?SFN=/this/is/dir1/file1.txt",
            },
            "/this/is/dir1/file2.foo.bar": {
                "SE1": "smr://srm.SE1.ch:8443/srm/v2/server?SFN=/this/is/dir1/file2.foo.bar",
                "SE3": "smr://srm.SE3.es:8443/srm/v2/server?SFN=/this/is/dir1/file2.foo.bar",
            },
        },
    },
}

fc_mock.getFileMetadata.return_value = S_OK(
    {"Successful": {"/a/lfn/1.txt": {"GUID": "AABB11"}, "/a/lfn/2.txt": {"GUID": "AABB22"}}, "Failed": {}}
)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from past.builtins import long
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities import ThreadScheduler
from DIRAC.Core.Base.ExecutorMindHandler import ExecutorMindHandler
from DIRAC.WorkloadManagementSystem.Client.JobState.JobState import JobState
from DIRAC.WorkloadManagementSystem.Client.JobState.CachedJobState import CachedJobState
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB
from DIRAC.WorkloadManagementSystem.DB.JobLoggingDB import JobLoggingDB
from DIRAC.WorkloadManagementSystem.DB.TaskQueueDB import TaskQueueDB


def cleanTaskQueues():
    tqDB = TaskQueueDB()
    jobDB = JobDB()
    logDB = JobLoggingDB()

    result = tqDB.enableAllTaskQueues()
    if not result["OK"]:
        return result
    result = tqDB.findOrphanJobs()
    if not result["OK"]:
        return result
    for jid in result["Value"]:
        result = tqDB.deleteJob(jid)
        if not result["OK"]:
            gLogger.error("Cannot delete from TQ job %s" % jid, result["Message"])
            continue
        result = jobDB.rescheduleJob(jid)
        if not result["OK"]:
            gLogger.error("Cannot reschedule in JobDB job %s" % jid, result["Message"])
            continue
        result = logDB.addLoggingRecord(jid, JobStatus.RECEIVED, "", "", source="JobState")
        if not result["OK"]:
            gLogger.error("Cannot add logging record in JobLoggingDB %s" % jid, result["Message"])
            continue
    return S_OK()


class OptimizationMindHandler(ExecutorMindHandler):

    __optimizationStates = [JobStatus.RECEIVED, JobStatus.CHECKING]
    __loadTaskId = False

    MSG_DEFINITIONS = {"OptimizeJobs": {"jids": (list, tuple)}}

    auth_msg_OptimizeJobs = ["all"]

    def msg_OptimizeJobs(self, msgObj):
        jids = msgObj.jids
        for jid in jids:
            try:
                jid = int(jid)
            except ValueError:
                self.log.error("Job ID %s has to be an integer" % jid)
                continue
            # Forget and add task to ensure state is reset
            self.forgetTask(jid)
            result = self.executeTask(jid, CachedJobState(jid))
            if not result["OK"]:
                self.log.error("Could not add job to optimization:", "%s %r" % (jid, result))
            else:
                self.log.info("Received new job", str(jid))
        return S_OK()

    @classmethod
    def __loadJobs(cls, eTypes=None):
        log = cls.log
        if cls.__loadTaskId:
            period = cls.srv_getCSOption("LoadJobPeriod", 300)
            ThreadScheduler.gThreadScheduler.setTaskPeriod(cls.__loadTaskId, period)
        if not eTypes:
            eConn = cls.getExecutorsConnected()
            eTypes = [eType for eType in eConn if eConn[eType] > 0]
        if not eTypes:
            log.info("No optimizer connected. Skipping load")
            return S_OK()
        log.info("Getting jobs for %s" % ",".join(eTypes))
        checkingMinors = [eType.split("/")[1] for eType in eTypes if eType != "WorkloadManagement/JobPath"]
        for opState in cls.__optimizationStates:
            # For Received states
            if opState == JobStatus.RECEIVED:
                if "WorkloadManagement/JobPath" not in eTypes:
                    continue
                jobCond = {"Status": opState}
            # For checking states
            if opState == JobStatus.CHECKING:
                if not checkingMinors:
                    continue
                jobCond = {"Status": opState, "MinorStatus": checkingMinors}
            # Do the magic
            jobTypeCondition = cls.srv_getCSOption("JobTypeRestriction", [])
            if jobTypeCondition:
                jobCond["JobType"] = jobTypeCondition
            result = cls.__jobDB.selectJobs(jobCond, limit=cls.srv_getCSOption("JobQueryLimit", 10000))
            if not result["OK"]:
                return result
            jidList = result["Value"]
            knownJids = cls.getTaskIds()
            added = 0
            for jid in jidList:
                jid = long(jid)
                if jid not in knownJids:
                    # Same as before. Check that the state is ok.
                    cls.executeTask(jid, CachedJobState(jid))
                    added += 1
            log.info("Added %s/%s jobs for %s state" % (added, len(jidList), opState))
        return S_OK()

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        try:
            from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB

            cls.__jobDB = JobDB()
        except Exception as excp:
            return S_ERROR("Could not connect to JobDB: %s" % str(excp))
        cls.setFailedOnTooFrozen(False)
        cls.setFreezeOnFailedDispatch(False)
        cls.setFreezeOnUnknownExecutor(False)
        cls.setAllowedClients("JobManager")
        cleanTaskQueues()
        period = cls.srv_getCSOption("LoadJobPeriod", 60)
        result = ThreadScheduler.gThreadScheduler.addPeriodicTask(period, cls.__loadJobs)
        if not result["OK"]:
            return result
        cls.__loadTaskId = result["Value"]
        return cls.__loadJobs()

    @classmethod
    def exec_executorConnected(cls, trid, eTypes):
        return cls.__loadJobs(eTypes)

    @classmethod
    def exec_taskProcessed(cls, jid, jobState, eType):
        cls.log.info("Saving changes for job", " %s after %s" % (jid, eType))
        result = jobState.commitChanges()
        if not result["OK"]:
            cls.log.error("Could not save changes for job", "%s: %s" % (jid, result["Message"]))
        return result

    @classmethod
    def exec_taskFreeze(cls, jid, jobState, eType):
        cls.log.info("Saving changes for job", " %s before freezing from %s" % (jid, eType))
        result = jobState.commitChanges()
        if not result["OK"]:
            cls.log.error("Could not save changes for job", "%s: %s" % (jid, result["Message"]))
        return result

    @classmethod
    def exec_dispatch(cls, jid, jobState, pathExecuted):
        result = jobState.getStatus()
        if not result["OK"]:
            cls.log.error("Could not get status for job", "%s: %s" % (jid, result["Message"]))
            return S_ERROR("Could not retrieve status: %s" % result["Message"])
        status, minorStatus = result["Value"]
        # If not in proper state then end chain
        if status not in cls.__optimizationStates:
            cls.log.info("Dispatching job %s out of optimization" % jid)
            return S_OK()
        # If received send to JobPath
        if status == JobStatus.RECEIVED:
            cls.log.info("Dispatching job", " %s to JobPath" % jid)
            return S_OK("WorkloadManagement/JobPath")
        result = jobState.getOptParameter("OptimizerChain")
        if not result["OK"]:
            cls.log.error(
                "Could not get optimizer chain for job, auto resetting job", "%s: %s" % (jid, result["Message"])
            )
            result = jobState.resetJob()
            if not result["OK"]:
                cls.log.error("Could not reset job", "%s: %s" % (jid, result["Message"]))
                return S_ERROR("Cound not get OptimizationChain or reset job %s" % jid)
            return S_OK("WorkloadManagement/JobPath")
        optChain = result["Value"]
        if minorStatus not in optChain:
            cls.log.error(
                "Next optimizer is not in the chain for job", "%s: %s not in %s" % (jid, minorStatus, optChain)
            )
            return S_ERROR("Next optimizer %s not in chain %s" % (minorStatus, optChain))
        cls.log.info("Dispatching job", " %s to %s" % (jid, minorStatus))
        return S_OK("WorkloadManagement/%s" % minorStatus)

    @classmethod
    def exec_prepareToSend(cls, jid, jobState, eId):
        return jobState.recheckValidity()

    @classmethod
    def exec_serializeTask(cls, jobState):
        return S_OK(jobState.serialize())

    @classmethod
    def exec_deserializeTask(cls, taskStub):
        return CachedJobState.deserialize(taskStub)

    @classmethod
    def exec_taskError(cls, jid, cachedJobState, errorMsg):
        if cachedJobState:
            result = cachedJobState.commitChanges()
            if not result["OK"]:
                cls.log.error("Cannot write changes to job %s: %s" % (jid, result["Message"]))
        else:
            cls.log.error(
                "Called exec_taskError with",
                "jid=%r cachedJobState=%r errorMsg=%r" % (jid, cachedJobState, errorMsg),
            )
        jobState = JobState(jid)
        result = jobState.getStatus()
        if not result["OK"]:
            cls.log.error("Could not get status of job %s: %s" % (jid, result["Message"]))
        elif result["Value"][0].lower() == "failed":
            return S_OK()
        cls.log.notice("Job %s: Setting to Failed|%s" % (jid, errorMsg))
        return jobState.setStatus("Failed", errorMsg, source="OptimizationMindHandler")
